__license__ = "GNU Affero General Public License http://www.gnu.org/licenses/agpl.html"
__copyright__ = "Copyright (C) 2014 The OctoPrint Project - Released under terms of the AGPLv3 License"

import hashlib
import logging
import os
import shutil
import time
import uuid

import wrapt
from flask_login import AnonymousUserMixin, UserMixin
from passlib.hash import pbkdf2_sha256
from werkzeug.local import LocalProxy

from octoprint.access.groups import Group, GroupChangeListener
from octoprint.access.permissions import OctoPrintPermission, Permissions
from octoprint.settings import settings as s
from octoprint.util import atomic_write, deprecated, generate_api_key
from octoprint.util import get_fully_qualified_classname as fqcn
from octoprint.util import to_bytes, yaml

password_hashers = []

try:
    from passlib.hash import argon2

    # test if we can actually hash and verify, if not we won't use this backend
    hash = argon2.hash("test")
    assert argon2.verify("test", hash)

    password_hashers.append(argon2)
except Exception:
    logging.getLogger(__name__).warning(
        "Argon2 passlib backend is not available, not using it for password hashing"
    )

password_hashers.append(pbkdf2_sha256)


class UserManager(GroupChangeListener):
    def __init__(self, group_manager, settings=None):
        self._group_manager = group_manager
        self._group_manager.register_listener(self)

        self._logger = logging.getLogger(__name__)
        self._session_users_by_session = {}
        self._sessionids_by_userid = {}

        if settings is None:
            settings = s()
        self._settings = settings

        self._login_status_listeners = []

    def register_login_status_listener(self, listener):
        self._login_status_listeners.append(listener)

    def unregister_login_status_listener(self, listener):
        self._login_status_listeners.remove(listener)

    def anonymous_user_factory(self):
        return AnonymousUser([self._group_manager.guest_group])

    def api_user_factory(self):
        return ApiUser([self._group_manager.admin_group, self._group_manager.user_group])

    @property
    def enabled(self):
        return True

    def login_user(self, user):
        self._cleanup_sessions()

        if user is None or user.is_anonymous:
            return

        if isinstance(user, LocalProxy):
            # noinspection PyProtectedMember
            user = user._get_current_object()

        if not isinstance(user, User):
            return None

        if not isinstance(user, SessionUser):
            user = SessionUser(user)

        self._session_users_by_session[user.session] = user

        userid = user.get_id()
        if userid not in self._sessionids_by_userid:
            self._sessionids_by_userid[userid] = set()

        self._sessionids_by_userid[userid].add(user.session)

        for listener in self._login_status_listeners:
            try:
                listener.on_user_logged_in(user)
            except Exception:
                self._logger.exception(
                    f"Error in on_user_logged_in on {listener!r}",
                    extra={"callback": fqcn(listener)},
                )

        self._logger.info(f"Logged in user: {user.get_id()}")

        return user

    def logout_user(self, user, stale=False):
        if user is None or user.is_anonymous:
            return

        if isinstance(user, LocalProxy):
            user = user._get_current_object()

        if not isinstance(user, SessionUser):
            return

        userid = user.get_id()
        sessionid = user.session

        if userid in self._sessionids_by_userid:
            try:
                self._sessionids_by_userid[userid].remove(sessionid)
            except KeyError:
                pass

        if sessionid in self._session_users_by_session:
            try:
                del self._session_users_by_session[sessionid]
            except KeyError:
                pass

        for listener in self._login_status_listeners:
            try:
                listener.on_user_logged_out(user, stale=stale)
            except Exception:
                self._logger.exception(
                    f"Error in on_user_logged_out on {listener!r}",
                    extra={"callback": fqcn(listener)},
                )

        self._logger.info(f"Logged out user: {user.get_id()}")

    def _cleanup_sessions(self):
        for session, user in list(self._session_users_by_session.items()):
            if not isinstance(user, SessionUser):
                continue
            if user.touched + (15 * 60) < time.monotonic():
                self._logger.info(
                    f"Cleaning up user session {session} for user {user.get_id()}"
                )
                self.logout_user(user, stale=True)

    @staticmethod
    def create_password_hash(password, *args, **kwargs):
        return password_hashers[0].hash(password)

    @staticmethod
    def create_legacy_password_hash(password, salt):
        return hashlib.sha512(
            to_bytes(password, encoding="utf-8", errors="replace") + to_bytes(salt)
        ).hexdigest()

    def check_password(self, username, password):
        user = self.find_user(username)
        if not user:
            return False

        if user.check_password(password):
            # password matches, correct password
            return True
        else:
            # new hash doesn't match, check legacy hash
            salt = self._settings.get(["accessControl"], asdict=True, merged=True).get(
                "salt", None
            )
            if salt is None:
                return False

            legacy_hash = UserManager.create_legacy_password_hash(password, salt)
            if user.check_password(legacy_hash, legacy=True):
                # legacy hash matches, we migrate the stored password hash to the new one and return True since it's the correct password
                self.change_user_password(username, password)
                return True
            else:
                # legacy hash doesn't match either, wrong password
                return False

    def cleanup_legacy_hashes(self):
        pass

    def signature_key_for_user(self, username, secret):
        return hashlib.sha512(
            to_bytes(username, encoding="utf-8", errors="replace") + to_bytes(secret)
        ).hexdigest()

    def add_user(self, username, password, active, permissions, groups, overwrite=False):
        pass

    def change_user_activation(self, username, active):
        pass

    def change_user_permissions(self, username, permissions):
        pass

    def add_permissions_to_user(self, username, permissions):
        pass

    def remove_permissions_from_user(self, username, permissions):
        pass

    def change_user_groups(self, username, groups):
        pass

    def add_groups_to_user(self, username, groups):
        pass

    def remove_groups_from_user(self, username, groups):
        pass

    def remove_groups_from_users(self, group):
        pass

    def change_user_password(self, username, password):
        pass

    def get_user_setting(self, username, key):
        return None

    def get_all_user_settings(self, username):
        return {}

    def change_user_setting(self, username, key, value):
        pass

    def change_user_settings(self, username, new_settings):
        pass

    def remove_user(self, username):
        if username in self._sessionids_by_userid:
            sessions = self._sessionids_by_userid[username]
            for session in sessions:
                if session in self._session_users_by_session:
                    del self._session_users_by_session[session]
            del self._sessionids_by_userid[username]

    def validate_user_session(self, userid, session):
        self._cleanup_sessions()

        if session in self._session_users_by_session:
            user = self._session_users_by_session[session]
            return userid == user.get_id()

        return False

    def find_user(self, userid=None, session=None, fresh=False):
        self._cleanup_sessions()

        if session is not None and session in self._session_users_by_session:
            user = self._session_users_by_session[session]
            if userid is None or userid == user.get_id():
                user.touch()
                return user

        return None

    def find_sessions_for(self, matcher):
        self._cleanup_sessions()

        result = []
        for user in self.get_all_users():
            if matcher(user):
                try:
                    session_ids = self._sessionids_by_userid[user.get_id()]
                    for session_id in session_ids:
                        try:
                            session_user = self._session_users_by_session[session_id]
                            session_user.touch()
                            result.append(session_user)
                        except KeyError:
                            # unknown session after all
                            continue
                except KeyError:
                    # no session for user
                    pass
        return result

    def get_all_users(self):
        return []

    def has_been_customized(self):
        return False

    def on_group_removed(self, group):
        self._logger.debug(f"Group {group.key} got removed, removing from all users")
        self.remove_groups_from_users([group])

    def on_group_permissions_changed(self, group, added=None, removed=None):
        users = self.find_sessions_for(lambda u: group in u.groups)
        for listener in self._login_status_listeners:
            try:
                for user in users:
                    listener.on_user_modified(user)
            except Exception:
                self._logger.exception(
                    f"Error in on_user_modified on {listener!r}",
                    extra={"callback": fqcn(listener)},
                )

    def on_group_subgroups_changed(self, group, added=None, removed=None):
        users = self.find_sessions_for(lambda u: group in u.groups)
        for listener in self._login_status_listeners:
            # noinspection PyBroadException
            try:
                for user in users:
                    listener.on_user_modified(user)
            except Exception:
                self._logger.exception(
                    f"Error in on_user_modified on {listener!r}",
                    extra={"callback": fqcn(listener)},
                )

    def _trigger_on_user_modified(self, user):
        if isinstance(user, str):
            # user id
            users = []
            try:
                session_ids = self._sessionids_by_userid[user]
                for session_id in session_ids:
                    try:
                        users.append(self._session_users_by_session[session_id])
                    except KeyError:
                        # unknown session id
                        continue
            except KeyError:
                # no session for user
                return
        elif isinstance(user, User) and not isinstance(user, SessionUser):
            users = self.find_sessions_for(lambda u: u.get_id() == user.get_id())
        elif isinstance(user, User):
            users = [user]
        else:
            return

        for listener in self._login_status_listeners:
            try:
                for user in users:
                    listener.on_user_modified(user)
            except Exception:
                self._logger.exception(
                    f"Error in on_user_modified on {listener!r}",
                    extra={"callback": fqcn(listener)},
                )

    def _trigger_on_user_removed(self, username):
        for listener in self._login_status_listeners:
            try:
                listener.on_user_removed(username)
            except Exception:
                self._logger.exception(
                    f"Error in on_user_removed on {listener!r}",
                    extra={"callback": fqcn(listener)},
                )

    # ~~ Deprecated methods follow

    # TODO: Remove deprecated methods in OctoPrint 1.5.0

    @deprecated(
        "changeUserRoles has been replaced by change_user_permissions",
        includedoc="Replaced by :func:`change_user_permissions`",
        since="1.4.0",
    )
    def changeUserRoles(self, username, roles):
        user = self.find_user(username)
        if user is None:
            raise UnknownUser(username)

        removed_roles = set(user._roles) - set(roles)
        self.removeRolesFromUser(username, removed_roles, user=user)

        added_roles = set(roles) - set(user._roles)
        self.addRolesToUser(username, added_roles, user=user)

    @deprecated(
        "addRolesToUser has been replaced by add_permissions_to_user",
        includedoc="Replaced by :func:`add_permissions_to_user`",
        since="1.4.0",
    )
    def addRolesToUser(self, username, roles, user=None):
        if user is None:
            user = self.find_user(username)

        if user is None:
            raise UnknownUser(username)

        if "admin" in roles:
            self.add_groups_to_user(username, self._group_manager.admin_group)

        if "user" in roles:
            self.remove_groups_from_user(username, self._group_manager.user_group)

    @deprecated(
        "removeRolesFromUser has been replaced by remove_permissions_from_user",
        includedoc="Replaced by :func:`remove_permissions_from_user`",
        since="1.4.0",
    )
    def removeRolesFromUser(self, username, roles, user=None):
        if user is None:
            user = self.find_user(username)

        if user is None:
            raise UnknownUser(username)

        if "admin" in roles:
            self.remove_groups_from_user(username, self._group_manager.admin_group)
            self.remove_permissions_from_user(username, Permissions.ADMIN)

        if "user" in roles:
            self.remove_groups_from_user(username, self._group_manager.user_group)

    checkPassword = deprecated(
        "checkPassword has been renamed to check_password",
        includedoc="Replaced by :func:`check_password`",
        since="1.4.0",
    )(check_password)
    addUser = deprecated(
        "addUser has been renamed to add_user",
        includedoc="Replaced by :func:`add_user`",
        since="1.4.0",
    )(add_user)
    changeUserActivation = deprecated(
        "changeUserActivation has been renamed to change_user_activation",
        includedoc="Replaced by :func:`change_user_activation`",
        since="1.4.0",
    )(change_user_activation)
    changeUserPassword = deprecated(
        "changeUserPassword has been renamed to change_user_password",
        includedoc="Replaced by :func:`change_user_password`",
        since="1.4.0",
    )(change_user_password)
    getUserSetting = deprecated(
        "getUserSetting has been renamed to get_user_setting",
        includedoc="Replaced by :func:`get_user_setting`",
        since="1.4.0",
    )(get_user_setting)
    getAllUserSettings = deprecated(
        "getAllUserSettings has been renamed to get_all_user_settings",
        includedoc="Replaced by :func:`get_all_user_settings`",
        since="1.4.0",
    )(get_all_user_settings)
    changeUserSetting = deprecated(
        "changeUserSetting has been renamed to change_user_setting",
        includedoc="Replaced by :func:`change_user_setting`",
        since="1.4.0",
    )(change_user_setting)
    changeUserSettings = deprecated(
        "changeUserSettings has been renamed to change_user_settings",
        includedoc="Replaced by :func:`change_user_settings`",
        since="1.4.0",
    )(change_user_settings)
    removeUser = deprecated(
        "removeUser has been renamed to remove_user",
        includedoc="Replaced by :func:`remove_user`",
        since="1.4.0",
    )(remove_user)
    findUser = deprecated(
        "findUser has been renamed to find_user",
        includedoc="Replaced by :func:`find_user`",
        since="1.4.0",
    )(find_user)
    getAllUsers = deprecated(
        "getAllUsers has been renamed to get_all_users",
        includedoc="Replaced by :func:`get_all_users`",
        since="1.4.0",
    )(get_all_users)
    hasBeenCustomized = deprecated(
        "hasBeenCustomized has been renamed to has_been_customized",
        includedoc="Replaced by :func:`has_been_customized`",
        since="1.4.0",
    )(has_been_customized)


class LoginStatusListener:
    def on_user_logged_in(self, user):
        pass

    def on_user_logged_out(self, user, stale=False):
        pass

    def on_user_modified(self, user):
        pass

    def on_user_removed(self, userid):
        pass


##~~ FilebasedUserManager, takes available users from users.yaml file


class FilebasedUserManager(UserManager):
    FILE_VERSION = 2

    def __init__(self, group_manager, path=None, settings=None):
        UserManager.__init__(self, group_manager, settings=settings)

        self._logger = logging.getLogger(__name__)

        if path is None:
            path = self._settings.get(["accessControl", "userfile"])
            if path is None:
                path = os.path.join(s().getBaseFolder("base"), "users.yaml")

        self._userfile = path

        self._users = {}
        self._dirty = False

        self._customized = None
        self._load()

    def _load(self):
        if os.path.exists(self._userfile) and os.path.isfile(self._userfile):
            data = yaml.load_from_file(path=self._userfile)

            if not data or not isinstance(data, dict):
                self._logger.fatal(
                    "{} does not contain a valid map of users. Fix "
                    "the file, or remove it, then restart OctoPrint.".format(
                        self._userfile
                    )
                )
                raise CorruptUserStorage()

            version = data.pop("_version", 1)
            if version != self.FILE_VERSION:
                self._logger.info(
                    f"Making a backup of the users.yaml file before migrating from version {version} to {self.FILE_VERSION}"
                )
                shutil.copy(
                    self._userfile,
                    os.path.splitext(self._userfile)[0] + f".v{version}.yaml",
                )
                self._dirty = True

            for name, attributes in data.items():
                if not isinstance(attributes, dict):
                    continue

                permissions = []
                if "permissions" in attributes:
                    permissions = attributes["permissions"]

                if "groups" in attributes:
                    groups = set(attributes["groups"])
                else:
                    groups = {self._group_manager.user_group}

                # migrate from roles to permissions
                if "roles" in attributes and "permissions" not in attributes:
                    self._logger.info(
                        f"Migrating user {name} to new granular permission system"
                    )

                    groups |= set(self._migrate_roles_to_groups(attributes["roles"]))
                    self._dirty = True

                apikey = None
                if "apikey" in attributes:
                    apikey = attributes["apikey"]
                settings = {}
                if "settings" in attributes:
                    settings = attributes["settings"]

                self._users[name] = User(
                    username=name,
                    passwordHash=attributes["password"],
                    active=attributes["active"],
                    permissions=self._to_permissions(*permissions),
                    groups=self._to_groups(*groups),
                    apikey=apikey,
                    settings=settings,
                )
                for sessionid in self._sessionids_by_userid.get(name, set()):
                    if sessionid in self._session_users_by_session:
                        self._session_users_by_session[sessionid].update_user(
                            self._users[name]
                        )

            if self._dirty:
                self._save()

            self.cleanup_legacy_hashes()

            self._customized = True
        else:
            self._customized = False

    def _save(self, force=False):
        if not self._dirty and not force:
            return

        data = {"_version": self.FILE_VERSION}
        for name, user in self._users.items():
            if not user or not isinstance(user, User):
                continue

            data[name] = {
                "password": user._passwordHash,
                "active": user._active,
                "groups": self._from_groups(*user._groups),
                "permissions": self._from_permissions(*user._permissions),
                "apikey": user._apikey,
                "settings": user._settings,
                # TODO: deprecated, remove in 1.5.0
                "roles": user._roles,
            }

        with atomic_write(
            self._userfile, mode="wt", permissions=0o600, max_permissions=0o666
        ) as f:
            yaml.save_to_file(data, file=f, pretty=True)
            self._dirty = False
        self._load()

    def _migrate_roles_to_groups(self, roles):
        # If admin is inside the roles, just return admin group
        if "admin" in roles:
            return [self._group_manager.admin_group, self._group_manager.user_group]
        else:
            return [self._group_manager.user_group]

    def _refresh_groups(self, user):
        user._groups = self._to_groups(*map(lambda g: g.key, user.groups))

    def add_user(
        self,
        username,
        password,
        active=False,
        permissions=None,
        groups=None,
        apikey=None,
        overwrite=False,
    ):
        if permissions is None:
            permissions = []
        permissions = self._to_permissions(*permissions)

        if groups is None:
            groups = self._group_manager.default_groups
        groups = self._to_groups(*groups)

        if username in self._users and not overwrite:
            raise UserAlreadyExists(username)

        if not username.strip() or username != username.strip():
            raise InvalidUsername(username)

        self._users[username] = User(
            username,
            UserManager.create_password_hash(password, settings=self._settings),
            active,
            permissions,
            groups,
            apikey=apikey,
        )
        self._dirty = True
        self._save()

    def change_user_activation(self, username, active):
        if username not in self._users:
            raise UnknownUser(username)

        if self._users[username].is_active != active:
            self._users[username]._active = active
            self._dirty = True
            self._save()

            self._trigger_on_user_modified(username)

    def change_user_permissions(self, username, permissions):
        if username not in self._users:
            raise UnknownUser(username)

        user = self._users[username]

        permissions = self._to_permissions(*permissions)

        removed_permissions = list(set(user._permissions) - set(permissions))
        added_permissions = list(set(permissions) - set(user._permissions))

        if len(removed_permissions) > 0:
            user.remove_permissions_from_user(removed_permissions)
            self._dirty = True

        if len(added_permissions) > 0:
            user.add_permissions_to_user(added_permissions)
            self._dirty = True

        if self._dirty:
            self._save()
            self._trigger_on_user_modified(username)

    def add_permissions_to_user(self, username, permissions):
        if username not in self._users:
            raise UnknownUser(username)

        if self._users[username].add_permissions_to_user(
            self._to_permissions(*permissions)
        ):
            self._dirty = True
            self._save()
            self._trigger_on_user_modified(username)

    def remove_permissions_from_user(self, username, permissions):
        if username not in self._users:
            raise UnknownUser(username)

        if self._users[username].remove_permissions_from_user(
            self._to_permissions(*permissions)
        ):
            self._dirty = True
            self._save()
            self._trigger_on_user_modified(username)

    def remove_permissions_from_users(self, permissions):
        modified = []
        for user in self._users:
            dirty = user.remove_permissions_from_user(self._to_permissions(*permissions))
            if dirty:
                self._dirty = True
                modified.append(user.get_id())

        if self._dirty:
            self._save()
            for username in modified:
                self._trigger_on_user_modified(username)

    def change_user_groups(self, username, groups):
        if username not in self._users:
            raise UnknownUser(username)

        user = self._users[username]

        groups = self._to_groups(*groups)

        removed_groups = list(set(user._groups) - set(groups))
        added_groups = list(set(groups) - set(user._groups))

        if len(removed_groups):
            self._dirty |= user.remove_groups_from_user(removed_groups)
        if len(added_groups):
            self._dirty |= user.add_groups_to_user(added_groups)

        if self._dirty:
            self._save()
            self._trigger_on_user_modified(username)

    def add_groups_to_user(self, username, groups, save=True, notify=True):
        if username not in self._users:
            raise UnknownUser(username)

        if self._users[username].add_groups_to_user(self._to_groups(*groups)):
            self._dirty = True

            if save:
                self._save()

            if notify:
                self._trigger_on_user_modified(username)

    def remove_groups_from_user(self, username, groups, save=True, notify=True):
        if username not in self._users:
            raise UnknownUser(username)

        if self._users[username].remove_groups_from_user(self._to_groups(*groups)):
            self._dirty = True

            if save:
                self._save()

            if notify:
                self._trigger_on_user_modified(username)

    def remove_groups_from_users(self, groups):
        modified = []
        for username, user in self._users.items():
            dirty = user.remove_groups_from_user(self._to_groups(*groups))
            if dirty:
                self._dirty = True
                modified.append(username)

        if self._dirty:
            self._save()

            for username in modified:
                self._trigger_on_user_modified(username)

    def change_user_password(self, username, password):
        if username not in self._users:
            raise UnknownUser(username)

        user = self._users[username]
        user._passwordHash = UserManager.create_password_hash(password)
        self._dirty = True
        self._save()

        self._trigger_on_user_modified(user)

    def cleanup_legacy_hashes(self):
        no_legacy = all(
            map(
                lambda u: u._passwordHash.startswith("$argon2id$")
                or u._passwordHash.startswith("$pbkdf2-"),
                self._users.values(),
            )
        )
        salt = self._settings.get(["accessControl"], asdict=True, merged=True).get(
            "salt", None
        )

        if no_legacy and salt:
            # no legacy hashes left, kill salt
            self._settings.backup("cleanup_legacy_hashes")
            self._settings.remove(["accessControl", "salt"])
            self._settings.save()

    def signature_key_for_user(self, username, secret):
        if username == "_api":
            return super().signature_key_for_user(username, secret)
        if username not in self._users:
            raise UnknownUser(username)
        user = self._users[username]

        return hashlib.sha512(
            to_bytes(username + user._passwordHash, encoding="utf-8", errors="replace")
            + to_bytes(secret)
        ).hexdigest()

    def change_user_setting(self, username, key, value):
        if username not in self._users:
            raise UnknownUser(username)

        user = self._users[username]
        old_value = user.get_setting(key)
        if not old_value or old_value != value:
            user.set_setting(key, value)
            self._dirty = self._dirty or old_value != value
            self._save()

    def change_user_settings(self, username, new_settings):
        if username not in self._users:
            raise UnknownUser(username)

        user = self._users[username]
        for key, value in new_settings.items():
            old_value = user.get_setting(key)
            user.set_setting(key, value)
            self._dirty = self._dirty or old_value != value
        self._save()

    def get_all_user_settings(self, username):
        if username not in self._users:
            raise UnknownUser(username)

        user = self._users[username]
        return user.get_all_settings()

    def get_user_setting(self, username, key):
        if username not in self._users:
            raise UnknownUser(username)

        user = self._users[username]
        return user.get_setting(key)

    def generate_api_key(self, username):
        if username not in self._users:
            raise UnknownUser(username)

        user = self._users[username]
        user._apikey = generate_api_key()
        self._dirty = True
        self._save()
        return user._apikey

    def delete_api_key(self, username):
        if username not in self._users:
            raise UnknownUser(username)

        user = self._users[username]
        user._apikey = None
        self._dirty = True
        self._save()

    def remove_user(self, username):
        UserManager.remove_user(self, username)

        if username not in self._users:
            raise UnknownUser(username)

        del self._users[username]
        self._dirty = True
        self._save()

    def find_user(self, userid=None, apikey=None, session=None, fresh=False):
        user = UserManager.find_user(self, userid=userid, session=session)

        if user is not None or (session and fresh):
            return user

        if userid is not None:
            if userid not in self._users:
                return None
            return self._users[userid]

        elif apikey is not None:
            for user in self._users.values():
                if apikey == user._apikey:
                    return user
            return None

        else:
            return None

    def get_all_users(self):
        return list(self._users.values())

    def has_been_customized(self):
        return self._customized

    def on_group_permissions_changed(self, group, added=None, removed=None):
        # refresh our group references
        for user in self.get_all_users():
            if group in user.groups:
                self._refresh_groups(user)

        # call parent
        UserManager.on_group_permissions_changed(
            self, group, added=added, removed=removed
        )

    def on_group_subgroups_changed(self, group, added=None, removed=None):
        # refresh our group references
        for user in self.get_all_users():
            if group in user.groups:
                self._refresh_groups(user)

        # call parent
        UserManager.on_group_subgroups_changed(self, group, added=added, removed=removed)

    # ~~ Helpers

    def _to_groups(self, *groups):
        return list(
            set(
                filter(
                    lambda x: x is not None,
                    (self._group_manager._to_group(group) for group in groups),
                )
            )
        )

    def _to_permissions(self, *permissions):
        return list(
            set(
                filter(
                    lambda x: x is not None,
                    (Permissions.find(permission) for permission in permissions),
                )
            )
        )

    def _from_groups(self, *groups):
        return list({group.key for group in groups})

    def _from_permissions(self, *permissions):
        return list({permission.key for permission in permissions})

    # ~~ Deprecated methods follow

    # TODO: Remove deprecated methods in OctoPrint 1.5.0

    generateApiKey = deprecated(
        "generateApiKey has been renamed to generate_api_key",
        includedoc="Replaced by :func:`generate_api_key`",
        since="1.4.0",
    )(generate_api_key)
    deleteApiKey = deprecated(
        "deleteApiKey has been renamed to delete_api_key",
        includedoc="Replaced by :func:`delete_api_key`",
        since="1.4.0",
    )(delete_api_key)
    addUser = deprecated(
        "addUser has been renamed to add_user",
        includedoc="Replaced by :func:`add_user`",
        since="1.4.0",
    )(add_user)
    changeUserActivation = deprecated(
        "changeUserActivation has been renamed to change_user_activation",
        includedoc="Replaced by :func:`change_user_activation`",
        since="1.4.0",
    )(change_user_activation)
    changeUserPassword = deprecated(
        "changeUserPassword has been renamed to change_user_password",
        includedoc="Replaced by :func:`change_user_password`",
        since="1.4.0",
    )(change_user_password)
    getUserSetting = deprecated(
        "getUserSetting has been renamed to get_user_setting",
        includedoc="Replaced by :func:`get_user_setting`",
        since="1.4.0",
    )(get_user_setting)
    getAllUserSettings = deprecated(
        "getAllUserSettings has been renamed to get_all_user_settings",
        includedoc="Replaced by :func:`get_all_user_settings`",
        since="1.4.0",
    )(get_all_user_settings)
    changeUserSetting = deprecated(
        "changeUserSetting has been renamed to change_user_setting",
        includedoc="Replaced by :func:`change_user_setting`",
        since="1.4.0",
    )(change_user_setting)
    changeUserSettings = deprecated(
        "changeUserSettings has been renamed to change_user_settings",
        includedoc="Replaced by :func:`change_user_settings`",
        since="1.4.0",
    )(change_user_settings)
    removeUser = deprecated(
        "removeUser has been renamed to remove_user",
        includedoc="Replaced by :func:`remove_user`",
        since="1.4.0",
    )(remove_user)
    findUser = deprecated(
        "findUser has been renamed to find_user",
        includedoc="Replaced by :func:`find_user`",
        since="1.4.0",
    )(find_user)
    getAllUsers = deprecated(
        "getAllUsers has been renamed to get_all_users",
        includedoc="Replaced by :func:`get_all_users`",
        since="1.4.0",
    )(get_all_users)
    hasBeenCustomized = deprecated(
        "hasBeenCustomized has been renamed to has_been_customized",
        includedoc="Replaced by :func:`has_been_customized`",
        since="1.4.0",
    )(has_been_customized)


##~~ Exceptions


class UserAlreadyExists(Exception):
    def __init__(self, username):
        Exception.__init__(self, "User %s already exists" % username)


class InvalidUsername(Exception):
    def __init__(self, username):
        Exception.__init__(self, "Username '%s' is invalid" % username)


class UnknownUser(Exception):
    def __init__(self, username):
        Exception.__init__(self, "Unknown user: %s" % username)


class UnknownRole(Exception):
    def __init__(self, role):
        Exception.__init__(self, "Unknown role: %s" % role)


class CorruptUserStorage(Exception):
    pass


##~~ Refactoring helpers


class MethodReplacedByBooleanProperty:
    def __init__(self, name, message, getter):
        self._name = name
        self._message = message
        self._getter = getter

    @property
    def _attr(self):
        return self._getter()

    def __call__(self):
        from warnings import warn

        warn(DeprecationWarning(self._message.format(name=self._name)), stacklevel=2)
        return self._attr

    def __eq__(self, other):
        return self._attr == other

    def __ne__(self, other):
        return self._attr != other

    def __bool__(self):
        # Python 3
        return self._attr

    def __nonzero__(self):
        # Python 2
        return self._attr

    def __hash__(self):
        return hash(self._attr)

    def __repr__(self):
        return "MethodReplacedByProperty({}, {}, {})".format(
            self._name, self._message, self._getter
        )

    def __str__(self):
        return str(self._attr)


# TODO: Remove compatibility layer in OctoPrint 1.5.0
class FlaskLoginMethodReplacedByBooleanProperty(MethodReplacedByBooleanProperty):
    def __init__(self, name, getter):
        message = (
            "{name} is now a property in Flask-Login versions >= 0.3.0, which OctoPrint now uses. "
            + "Use {name} instead of {name}(). This compatibility layer will be removed in OctoPrint 1.5.0."
        )
        MethodReplacedByBooleanProperty.__init__(self, name, message, getter)


# TODO: Remove compatibility layer in OctoPrint 1.5.0
class OctoPrintUserMethodReplacedByBooleanProperty(MethodReplacedByBooleanProperty):
    def __init__(self, name, getter):
        message = (
            "{name} is now a property for consistency reasons with Flask-Login versions >= 0.3.0, which "
            + "OctoPrint now uses. Use {name} instead of {name}(). This compatibility layer will be removed "
            + "in OctoPrint 1.5.0."
        )
        MethodReplacedByBooleanProperty.__init__(self, name, message, getter)


##~~ User object


class User(UserMixin):
    def __init__(
        self,
        username,
        passwordHash,
        active,
        permissions=None,
        groups=None,
        apikey=None,
        settings=None,
    ):
        if permissions is None:
            permissions = []
        if groups is None:
            groups = []

        self._username = username
        self._passwordHash = passwordHash
        self._active = active
        self._permissions = permissions
        self._groups = groups
        self._apikey = apikey

        if settings is None:
            settings = {}

        self._settings = settings

    def as_dict(self):
        from octoprint.access.permissions import OctoPrintPermission

        return {
            "name": self._username,
            "active": bool(self.is_active),
            "permissions": list(map(lambda p: p.key, self._permissions)),
            "groups": list(map(lambda g: g.key, self._groups)),
            "needs": OctoPrintPermission.convert_needs_to_dict(self.needs),
            "apikey": self._apikey,
            "settings": self._settings,
            # TODO: deprecated, remove in 1.5.0
            "admin": self.has_permission(Permissions.ADMIN),
            "user": not self.is_anonymous,
            "roles": self._roles,
        }

    def check_password(self, password, legacy=False):
        if legacy:
            return self._passwordHash == password

        for password_hash in password_hashers:
            if password_hash.identify(self._passwordHash):
                try:
                    return password_hash.verify(password, self._passwordHash)
                except ValueError:
                    pass

        return False

    def get_id(self):
        return self.get_name()

    def get_name(self):
        return self._username

    @property
    def is_anonymous(self):
        return FlaskLoginMethodReplacedByBooleanProperty("is_anonymous", lambda: False)

    @property
    def is_authenticated(self):
        return FlaskLoginMethodReplacedByBooleanProperty("is_authenticated", lambda: True)

    @property
    def is_active(self):
        return FlaskLoginMethodReplacedByBooleanProperty(
            "is_active", lambda: self._active
        )

    def get_all_settings(self):
        return self._settings

    def get_setting(self, key):
        if not isinstance(key, (tuple, list)):
            path = [key]
        else:
            path = key

        return self._get_setting(path)

    def set_setting(self, key, value):
        if not isinstance(key, (tuple, list)):
            path = [key]
        else:
            path = key
        return self._set_setting(path, value)

    def _get_setting(self, path):
        s = self._settings
        for p in path:
            if isinstance(s, dict) and p in s:
                s = s[p]
            else:
                return None
        return s

    def _set_setting(self, path, value):
        s = self._settings
        for p in path[:-1]:
            if p not in s:
                s[p] = {}

            if not isinstance(s[p], dict):
                s[p] = {}

            s = s[p]

        key = path[-1]
        s[key] = value
        return True

    def add_permissions_to_user(self, permissions):
        # Make sure the permissions variable is of type list
        if not isinstance(permissions, list):
            permissions = [permissions]

        assert all(map(lambda p: isinstance(p, OctoPrintPermission), permissions))

        dirty = False
        for permission in permissions:
            if permissions not in self._permissions:
                self._permissions.append(permission)
                dirty = True

        return dirty

    def remove_permissions_from_user(self, permissions):
        # Make sure the permissions variable is of type list
        if not isinstance(permissions, list):
            permissions = [permissions]

        assert all(map(lambda p: isinstance(p, OctoPrintPermission), permissions))

        dirty = False
        for permission in permissions:
            if permission in self._permissions:
                self._permissions.remove(permission)
                dirty = True

        return dirty

    def add_groups_to_user(self, groups):
        # Make sure the groups variable is of type list
        if not isinstance(groups, list):
            groups = [groups]

        assert all(map(lambda p: isinstance(p, Group), groups))

        dirty = False
        for group in groups:
            if group.is_toggleable() and group not in self._groups:
                self._groups.append(group)
                dirty = True

        return dirty

    def remove_groups_from_user(self, groups):
        # Make sure the groups variable is of type list
        if not isinstance(groups, list):
            groups = [groups]

        assert all(map(lambda p: isinstance(p, Group), groups))

        dirty = False
        for group in groups:
            if group.is_toggleable() and group in self._groups:
                self._groups.remove(group)
                dirty = True

        return dirty

    @property
    def permissions(self):
        if self._permissions is None:
            return []

        if Permissions.ADMIN in self._permissions:
            return Permissions.all()

        return list(filter(lambda p: p is not None, self._permissions))

    @property
    def groups(self):
        return list(self._groups)

    @property
    def effective_permissions(self):
        if self._permissions is None:
            return []
        return list(
            filter(lambda p: p is not None and self.has_permission(p), Permissions.all())
        )

    @property
    def needs(self):
        needs = set()

        for permission in self.permissions:
            if permission is not None:
                needs = needs.union(permission.needs)

        for group in self.groups:
            if group is not None:
                needs = needs.union(group.needs)

        return needs

    def has_permission(self, permission):
        return self.has_needs(*permission.needs)

    def has_needs(self, *needs):
        return set(needs).issubset(self.needs)

    def __repr__(self):
        return (
            "User(id=%s,name=%s,active=%r,user=True,admin=%r,permissions=%s,groups=%s)"
            % (
                self.get_id(),
                self.get_name(),
                bool(self.is_active),
                self.has_permission(Permissions.ADMIN),
                self._permissions,
                self._groups,
            )
        )

    # ~~ Deprecated methods & properties follow

    # TODO: Remove deprecated methods & properties in OctoPrint 1.5.0

    asDict = deprecated(
        "asDict has been renamed to as_dict",
        includedoc="Replaced by :func:`as_dict`",
        since="1.4.0",
    )(as_dict)

    @property
    @deprecated("is_user is deprecated, please use has_permission", since="1.4.0")
    def is_user(self):
        return OctoPrintUserMethodReplacedByBooleanProperty(
            "is_user", lambda: not self.is_anonymous
        )

    @property
    @deprecated("is_admin is deprecated, please use has_permission", since="1.4.0")
    def is_admin(self):
        return OctoPrintUserMethodReplacedByBooleanProperty(
            "is_admin", lambda: self.has_permission(Permissions.ADMIN)
        )

    @property
    @deprecated("roles is deprecated, please use has_permission", since="1.4.0")
    def roles(self):
        return self._roles

    @property
    def _roles(self):
        """Helper for the deprecated self.roles and serializing to yaml"""
        if self.has_permission(Permissions.ADMIN):
            return ["user", "admin"]
        elif not self.is_anonymous:
            return ["user"]
        else:
            return []


class AnonymousUser(AnonymousUserMixin, User):
    def __init__(self, groups):
        User.__init__(self, None, "", True, [], groups)

    @property
    def is_anonymous(self):
        return FlaskLoginMethodReplacedByBooleanProperty("is_anonymous", lambda: True)

    @property
    def is_authenticated(self):
        return FlaskLoginMethodReplacedByBooleanProperty(
            "is_authenticated", lambda: False
        )

    @property
    def is_active(self):
        return FlaskLoginMethodReplacedByBooleanProperty(
            "is_active", lambda: self._active
        )

    def check_password(self, passwordHash):
        return True

    def as_dict(self):
        from octoprint.access.permissions import OctoPrintPermission

        return {"needs": OctoPrintPermission.convert_needs_to_dict(self.needs)}

    def __repr__(self):
        return "AnonymousUser(groups=%s)" % self._groups


class SessionUser(wrapt.ObjectProxy):
    def __init__(self, user):
        wrapt.ObjectProxy.__init__(self, user)

        self._self_session = "".join("%02X" % z for z in bytes(uuid.uuid4().bytes))
        self._self_created = self._self_touched = time.monotonic()

    @property
    def session(self):
        return self._self_session

    @property
    def created(self):
        return self._self_created

    @property
    def touched(self):
        return self._self_touched

    def touch(self):
        self._self_touched = time.monotonic()

    @deprecated(
        "SessionUser.get_session() has been deprecated, use SessionUser.session instead",
        since="1.3.5",
    )
    def get_session(self):
        return self.session

    def update_user(self, user):
        self.__wrapped__ = user

    def as_dict(self):
        result = self.__wrapped__.as_dict()
        result.update({"session": self.session})
        return result

    def __repr__(self):
        return "SessionUser({!r},session={},created={})".format(
            self.__wrapped__, self.session, self.created
        )


##~~ User object to use when global api key is used to access the API


class ApiUser(User):
    def __init__(self, groups):
        User.__init__(self, "_api", "", True, [], groups)



__license__ = "GNU Affero General Public License http://www.gnu.org/licenses/agpl.html"
__copyright__ = "Copyright (C) 2022 The OctoPrint Project - Released under terms of the AGPLv3 License"

from typing import Dict, Optional

from octoprint.schema import BaseModel
from octoprint.vendor.with_attrs_docs import with_attrs_docs


@with_attrs_docs
class ApiConfig(BaseModel):
    key: Optional[str] = None
    """Global API key, deprecated, use User API keys instead. Unset by default, will be generated on first run."""

    apps: Dict[str, str] = {}

    allowCrossOrigin: bool = False
    """Whether to allow cross origin access to the API or not."""



"""
This module contains printer profile related code.

A printer profile is a ``dict`` of the following structure:

.. list-table::
   :widths: 15 5 30
   :header-rows: 1

   * - Name
     - Type
     - Description
   * - ``id``
     - ``string``
     - Internal id of the printer profile
   * - ``name``
     - ``string``
     - Human readable name of the printer profile
   * - ``model``
     - ``string``
     - Printer model
   * - ``color``
     - ``string``
     - Color to associate with the printer profile
   * - ``volume``
     - ``dict``
     - Information about the print volume
   * - ``volume.width``
     - ``float``
     - Width of the print volume (X axis)
   * - ``volume.depth``
     - ``float``
     - Depth of the print volume (Y axis)
   * - ``volume.height``
     - ``float``
     - Height of the print volume (Z axis)
   * - ``volume.formFactor``
     - ``string``
     - Form factor of the print bed, either ``rectangular`` or ``circular``
   * - ``volume.origin``
     - ``string``
     - Location of gcode origin in the print volume, either ``lowerleft`` or ``center``
   * - ``volume.custom_box``
     - ``dict`` or ``False``
     - Custom boundary box overriding the default bounding box based on the provided width, depth, height and origin.
       If ``False``, the default boundary box will be used.
   * - ``volume.custom_box.x_min``
     - ``float``
     - Minimum valid X coordinate
   * - ``volume.custom_box.y_min``
     - ``float``
     - Minimum valid Y coordinate
   * - ``volume.custom_box.z_min``
     - ``float``
     - Minimum valid Z coordinate
   * - ``volume.custom_box.x_max``
     - ``float``
     - Maximum valid X coordinate
   * - ``volume.custom_box.y_max``
     - ``float``
     - Maximum valid Y coordinate
   * - ``volume.custom_box.z_max``
     - ``float``
     - Maximum valid Z coordinate
   * - ``heatedBed``
     - ``bool``
     - Whether the printer has a heated bed (``True``) or not (``False``)
   * - ``heatedChamber``
     - ``bool``
     - Whether the printer has a heated chamber (``True``) or not (``False``)
   * - ``extruder``
     - ``dict``
     - Information about the printer's extruders
   * - ``extruder.count``
     - ``int``
     - How many extruders the printer has (default 1)
   * - ``extruder.offsets``
     - ``list`` of ``tuple``
     - Extruder offsets relative to first extruder, list of (x, y) tuples, first is always (0,0)
   * - ``extruder.nozzleDiameter``
     - ``float``
     - Diameter of the printer nozzle(s)
   * - ``extruder.sharedNozzle``
     - ``boolean``
     - Whether there's only one nozzle shared among all extruders (true) or one nozzle per extruder (false).
   * - ``extruder.defaultExtrusionLength``
     - ``int``
     - Default extrusion length used in Control tab on initial page load in mm.
   * - ``axes``
     - ``dict``
     - Information about the printer axes
   * - ``axes.x``
     - ``dict``
     - Information about the printer's X axis
   * - ``axes.x.speed``
     - ``float``
     - Speed of the X axis in mm/min
   * - ``axes.x.inverted``
     - ``bool``
     - Whether a positive value change moves the nozzle away from the print bed's origin (False, default) or towards it (True)
   * - ``axes.y``
     - ``dict``
     - Information about the printer's Y axis
   * - ``axes.y.speed``
     - ``float``
     - Speed of the Y axis in mm/min
   * - ``axes.y.inverted``
     - ``bool``
     - Whether a positive value change moves the nozzle away from the print bed's origin (False, default) or towards it (True)
   * - ``axes.z``
     - ``dict``
     - Information about the printer's Z axis
   * - ``axes.z.speed``
     - ``float``
     - Speed of the Z axis in mm/min
   * - ``axes.z.inverted``
     - ``bool``
     - Whether a positive value change moves the nozzle away from the print bed (False, default) or towards it (True)
   * - ``axes.e``
     - ``dict``
     - Information about the printer's E axis
   * - ``axes.e.speed``
     - ``float``
     - Speed of the E axis in mm/min
   * - ``axes.e.inverted``
     - ``bool``
     - Whether a positive value change extrudes (False, default) or retracts (True) filament

.. autoclass:: PrinterProfileManager
   :members:

.. autoclass:: BedFormFactor
   :members:

.. autoclass:: BedOrigin
   :members:

.. autoclass:: SaveError

.. autoclass:: CouldNotOverwriteError

.. autoclass:: InvalidProfileError
"""

__author__ = "Gina Häußge <osd@foosel.net>"
__license__ = "GNU Affero General Public License http://www.gnu.org/licenses/agpl.html"
__copyright__ = "Copyright (C) 2014 The OctoPrint Project - Released under terms of the AGPLv3 License"


import copy
import logging
import os

from octoprint.settings import settings
from octoprint.util import (
    dict_contains_keys,
    dict_merge,
    dict_sanitize,
    is_hidden_path,
    yaml,
)


class SaveError(Exception):
    """Saving a profile failed"""

    pass


class CouldNotOverwriteError(SaveError):
    """Overwriting of a profile not allowed"""

    pass


class InvalidProfileError(Exception):
    """Profile invalid"""

    pass


class BedFormFactor:
    """Valid values for bed form factor"""

    RECTANGULAR = "rectangular"
    """Rectangular bed"""

    CIRCULAR = "circular"
    """Circular bed"""

    @classmethod
    def values(cls):
        return [
            getattr(cls, name)
            for name in cls.__dict__
            if not (name.startswith("__") or name == "values")
        ]


BedTypes = BedFormFactor
"""Deprecated name of :class:`BedFormFactors`"""


class BedOrigin:
    """Valid values for bed origin"""

    LOWERLEFT = "lowerleft"
    """Origin at lower left corner of the bed when looking from the top"""

    CENTER = "center"
    """Origin at the center of the bed when looking from top"""

    @classmethod
    def values(cls):
        return [
            getattr(cls, name)
            for name in cls.__dict__
            if not (name.startswith("__") or name == "values")
        ]


class PrinterProfileManager:
    """
    Manager for printer profiles. Offers methods to select the globally used printer profile and to list, add, remove,
    load and save printer profiles.
    """

    default = {
        "id": "_default",
        "name": "Default",
        "model": "Generic RepRap Printer",
        "color": "default",
        "volume": {
            "width": 200,
            "depth": 200,
            "height": 200,
            "formFactor": BedFormFactor.RECTANGULAR,
            "origin": BedOrigin.LOWERLEFT,
            "custom_box": False,
        },
        "heatedBed": True,
        "heatedChamber": False,
        "extruder": {
            "count": 1,
            "offsets": [(0, 0)],
            "nozzleDiameter": 0.4,
            "sharedNozzle": False,
            "defaultExtrusionLength": 5,
        },
        "axes": {
            "x": {"speed": 6000, "inverted": False},
            "y": {"speed": 6000, "inverted": False},
            "z": {"speed": 200, "inverted": False},
            "e": {"speed": 300, "inverted": False},
        },
    }

    def __init__(self):
        self._current = None
        self._logger = logging.getLogger(__name__)

        fresh = settings().checkBaseFolder("printerProfiles")
        self._folder = settings().getBaseFolder("printerProfiles")

        self._migrate_old_default_profile()
        self._verify_default_available(fresh=fresh)

    def _migrate_old_default_profile(self):
        default_overrides = settings().get(["printerProfiles", "defaultProfile"])
        if not default_overrides:
            return

        if self.exists("_default"):
            return

        if not isinstance(default_overrides, dict):
            self._logger.warning(
                "Existing default profile from settings is not a valid profile, refusing to migrate: {!r}".format(
                    default_overrides
                )
            )
            return

        default_overrides["id"] = "_default"
        self.save(default_overrides)

        settings().set(["printerProfiles", "defaultProfile"], None)
        settings().save()

        self._logger.info(
            "Migrated default printer profile from settings to _default.profile: {!r}".format(
                default_overrides
            )
        )

    def _verify_default_available(self, fresh=False):
        default_id = settings().get(["printerProfiles", "default"])
        if default_id is None:
            default_id = "_default"

        if not self.exists(default_id):
            if not self.exists("_default"):
                if not fresh:
                    if default_id == "_default":
                        self._logger.error(
                            "Profile _default does not exist, creating _default again and setting it as default"
                        )
                    else:
                        self._logger.error(
                            "Selected default profile {} and _default do not exist, creating _default again and setting it as default".format(
                                default_id
                            )
                        )
                self.save(self.__class__.default, allow_overwrite=True, make_default=True)
            else:
                self._logger.error(
                    "Selected default profile {} does not exists, resetting to _default".format(
                        default_id
                    )
                )
                settings().set(["printerProfiles", "default"], "_default")
                settings().save()
            default_id = "_default"

        profile = self.get(default_id)
        if profile is None:
            self._logger.error(
                "Selected default profile {} is invalid, resetting to default values".format(
                    default_id
                )
            )
            profile = copy.deepcopy(self.__class__.default)
            profile["id"] = default_id
            self.save(self.__class__.default, allow_overwrite=True, make_default=True)

    def select(self, identifier):
        if identifier is None or not self.exists(identifier):
            self._current = self.get_default()
            return False
        else:
            self._current = self.get(identifier)
            if self._current is None:
                self._logger.error(
                    "Profile {} is invalid, cannot select, falling back to default".format(
                        identifier
                    )
                )
                self._current = self.get_default()
                return False
            else:
                return True

    def deselect(self):
        self._current = None

    def get_all(self):
        return self._load_all()

    def get(self, identifier):
        try:
            if self.exists(identifier):
                return self._load_from_path(self._get_profile_path(identifier))
            else:
                return None
        except InvalidProfileError:
            return None

    def remove(self, identifier, trigger_event=False):
        if self._current is not None and self._current["id"] == identifier:
            return False
        elif settings().get(["printerProfiles", "default"]) == identifier:
            return False

        removed = self._remove_from_path(self._get_profile_path(identifier))
        if removed and trigger_event:
            from octoprint.events import Events, eventManager

            payload = {"identifier": identifier}
            eventManager().fire(Events.PRINTER_PROFILE_DELETED, payload=payload)

        return removed

    def save(
        self, profile, allow_overwrite=False, make_default=False, trigger_event=False
    ):
        if "id" in profile:
            identifier = profile["id"]
        elif "name" in profile:
            identifier = profile["name"]
        else:
            raise InvalidProfileError("profile must contain either id or name")

        identifier = self._sanitize(identifier)
        profile["id"] = identifier

        self._migrate_profile(profile)
        profile = dict_sanitize(profile, self.__class__.default)
        profile = dict_merge(self.__class__.default, profile)

        path = self._get_profile_path(identifier)
        is_overwrite = os.path.exists(path)

        self._save_to_path(path, profile, allow_overwrite=allow_overwrite)

        if make_default:
            settings().set(["printerProfiles", "default"], identifier)
            settings().save()

        if self._current is not None and self._current["id"] == identifier:
            self.select(identifier)

        from octoprint.events import Events, eventManager

        if trigger_event:
            payload = {"identifier": identifier}
            event = (
                Events.PRINTER_PROFILE_MODIFIED
                if is_overwrite
                else Events.PRINTER_PROFILE_ADDED
            )
            eventManager().fire(event, payload=payload)

        return self.get(identifier)

    def is_default_unmodified(self):
        default = settings().get(["printerProfiles", "default"])
        return default is None or default == "_default" or not self.exists("_default")

    @property
    def profile_count(self):
        return len(self._load_all_identifiers())

    @property
    def last_modified(self):
        dates = [os.stat(self._folder).st_mtime]
        dates += [
            entry.stat().st_mtime
            for entry in os.scandir(self._folder)
            if entry.name.endswith(".profile")
        ]
        return max(dates)

    def get_default(self):
        default = settings().get(["printerProfiles", "default"])
        if default is not None and self.exists(default):
            profile = self.get(default)
            if profile is not None:
                return profile
            else:
                self._logger.warning(
                    "Default profile {} is invalid, falling back to built-in defaults".format(
                        default
                    )
                )

        return copy.deepcopy(self.__class__.default)

    def set_default(self, identifier):
        all_identifiers = self._load_all_identifiers()
        if identifier is not None and identifier not in all_identifiers:
            return

        settings().set(["printerProfiles", "default"], identifier)
        settings().save()

    def get_current_or_default(self):
        if self._current is not None:
            return self._current
        else:
            return self.get_default()

    def get_current(self):
        return self._current

    def exists(self, identifier):
        if identifier is None:
            return False
        else:
            path = self._get_profile_path(identifier)
            return os.path.exists(path) and os.path.isfile(path)

    def _load_all(self):
        all_identifiers = self._load_all_identifiers()
        results = {}
        for identifier, path in all_identifiers.items():
            try:
                profile = self._load_from_path(path)
            except InvalidProfileError:
                self._logger.warning(f"Profile {identifier} is invalid, skipping")
                continue

            if profile is None:
                continue

            results[identifier] = dict_merge(self.__class__.default, profile)
        return results

    def _load_all_identifiers(self):
        results = {}
        for entry in os.scandir(self._folder):
            if is_hidden_path(entry.name) or not entry.name.endswith(".profile"):
                continue

            if not entry.is_file():
                continue

            identifier = entry.name[: -len(".profile")]
            results[identifier] = entry.path
        return results

    def _load_from_path(self, path):
        if not os.path.exists(path) or not os.path.isfile(path):
            return None

        profile = yaml.load_from_file(path=path)

        if profile is None or not isinstance(profile, dict):
            raise InvalidProfileError("Profile is None or not a dictionary")

        if self._migrate_profile(profile):
            try:
                self._save_to_path(path, profile, allow_overwrite=True)
            except Exception:
                self._logger.exception(
                    "Tried to save profile to {path} after migrating it while loading, ran into exception".format(
                        path=path
                    )
                )

        profile = self._ensure_valid_profile(profile)

        if not profile:
            self._logger.warning("Invalid profile: %s" % path)
            raise InvalidProfileError()
        return profile

    def _save_to_path(self, path, profile, allow_overwrite=False):
        validated_profile = self._ensure_valid_profile(profile)
        if not validated_profile:
            raise InvalidProfileError()

        if os.path.exists(path) and not allow_overwrite:
            raise SaveError(
                "Profile %s already exists and not allowed to overwrite" % profile["id"]
            )

        from octoprint.util import atomic_write

        try:
            with atomic_write(path, mode="wt", max_permissions=0o666) as f:
                yaml.save_to_file(profile, file=f, pretty=True)
        except Exception as e:
            self._logger.exception(
                "Error while trying to save profile %s" % profile["id"]
            )
            raise SaveError("Cannot save profile {}: {}".format(profile["id"], str(e)))

    def _remove_from_path(self, path):
        try:
            os.remove(path)
            return True
        except Exception:
            return False

    def _get_profile_path(self, identifier):
        return os.path.join(self._folder, "%s.profile" % identifier)

    def _sanitize(self, name):
        if name is None:
            return None

        if "/" in name or "\\" in name:
            raise ValueError("name must not contain / or \\")

        import string

        valid_chars = "-_.() {ascii}{digits}".format(
            ascii=string.ascii_letters, digits=string.digits
        )
        sanitized_name = "".join(c for c in name if c in valid_chars)
        sanitized_name = sanitized_name.replace(" ", "_")
        return sanitized_name

    def _migrate_profile(self, profile):
        # make sure profile format is up to date
        modified = False

        if (
            "volume" in profile
            and "formFactor" in profile["volume"]
            and "origin" not in profile["volume"]
        ):
            profile["volume"]["origin"] = (
                BedOrigin.CENTER
                if profile["volume"]["formFactor"] == BedFormFactor.CIRCULAR
                else BedOrigin.LOWERLEFT
            )
            modified = True

        if "volume" in profile and "custom_box" not in profile["volume"]:
            profile["volume"]["custom_box"] = False
            modified = True

        if "extruder" in profile and "sharedNozzle" not in profile["extruder"]:
            profile["extruder"]["sharedNozzle"] = False
            modified = True

        if (
            "extruder" in profile
            and "sharedNozzle" in profile["extruder"]
            and profile["extruder"]["sharedNozzle"]
        ):
            profile["extruder"]["offsets"] = [(0.0, 0.0)]
            modified = True

        if "extruder" in profile and "defaultExtrusionLength" not in profile["extruder"]:
            profile["extruder"]["defaultExtrusionLength"] = self.default["extruder"][
                "defaultExtrusionLength"
            ]
            modified = True

        if "heatedChamber" not in profile:
            profile["heatedChamber"] = False
            modified = True

        return modified

    def _ensure_valid_profile(self, profile):
        # ensure all keys are present
        if not dict_contains_keys(self.default, profile):
            self._logger.warning(
                "Profile invalid, missing keys. Expected: {expected!r}. Actual: {actual!r}".format(
                    expected=list(self.default.keys()), actual=list(profile.keys())
                )
            )
            return False

        # conversion helper
        def convert_value(profile, path, converter):
            value = profile
            for part in path[:-1]:
                if not isinstance(value, dict) or part not in value:
                    raise RuntimeError("%s is not contained in profile" % ".".join(path))
                value = value[part]

            if not isinstance(value, dict) or path[-1] not in value:
                raise RuntimeError("%s is not contained in profile" % ".".join(path))

            value[path[-1]] = converter(value[path[-1]])

        def validate_value(profile, path, validator):
            value = profile
            for part in path[:-1]:
                if not isinstance(value, dict) or part not in value:
                    raise RuntimeError("%s is not contained in profile" % ".".join(path))
                value = value[part]

            if not isinstance(value, dict) or path[-1] not in value:
                raise RuntimeError("%s is not contained in profile" % ".".join(path))

            return validator(value[path[-1]])

        # convert ints
        for path in (
            ("extruder", "count"),
            ("axes", "x", "speed"),
            ("axes", "y", "speed"),
            ("axes", "z", "speed"),
            ("axes", "e", "speed"),
        ):
            try:
                convert_value(profile, path, int)
            except Exception as e:
                self._logger.warning(
                    "Profile has invalid value for path {path!r}: {msg}".format(
                        path=".".join(path), msg=str(e)
                    )
                )
                return False

        # convert floats
        for path in (
            ("volume", "width"),
            ("volume", "depth"),
            ("volume", "height"),
            ("extruder", "nozzleDiameter"),
        ):
            try:
                convert_value(profile, path, float)
            except Exception as e:
                self._logger.warning(
                    "Profile has invalid value for path {path!r}: {msg}".format(
                        path=".".join(path), msg=str(e)
                    )
                )
                return False

        # convert booleans
        for path in (
            ("axes", "x", "inverted"),
            ("axes", "y", "inverted"),
            ("axes", "z", "inverted"),
            ("extruder", "sharedNozzle"),
        ):
            try:
                convert_value(profile, path, bool)
            except Exception as e:
                self._logger.warning(
                    "Profile has invalid value for path {path!r}: {msg}".format(
                        path=".".join(path), msg=str(e)
                    )
                )
                return False

        # validate volume size range
        for path in (
            ("volume", "width"),
            ("volume", "depth"),
            ("volume", "height"),
        ):
            if not validate_value(profile, path, lambda x: x > 0 and x < 10000):
                return False

        # validate extruder count range
        if not validate_value(
            profile, ("extruder", "count"), lambda x: x > 0 and x < 100
        ):
            return False

        # validate form factor
        if profile["volume"]["formFactor"] not in BedFormFactor.values():
            self._logger.warning(
                "Profile has invalid value volume.formFactor: {formFactor}".format(
                    formFactor=profile["volume"]["formFactor"]
                )
            )
            return False

        # validate origin type
        if profile["volume"]["origin"] not in BedOrigin.values():
            self._logger.warning(
                "Profile has invalid value in volume.origin: {origin}".format(
                    origin=profile["volume"]["origin"]
                )
            )
            return False

        # ensure origin and form factor combination is legal
        if (
            profile["volume"]["formFactor"] == BedFormFactor.CIRCULAR
            and not profile["volume"]["origin"] == BedOrigin.CENTER
        ):
            profile["volume"]["origin"] = BedOrigin.CENTER

        # force width and depth of volume to be identical for circular beds, with width being the reference
        if profile["volume"]["formFactor"] == BedFormFactor.CIRCULAR:
            profile["volume"]["depth"] = profile["volume"]["width"]

        # if we have a custom bounding box, validate it
        if profile["volume"]["custom_box"] and isinstance(
            profile["volume"]["custom_box"], dict
        ):
            if not len(profile["volume"]["custom_box"]):
                profile["volume"]["custom_box"] = False

            else:
                default_box = self._default_box_for_volume(profile["volume"])
                for prop, limiter in (
                    ("x_min", min),
                    ("y_min", min),
                    ("z_min", min),
                    ("x_max", max),
                    ("y_max", max),
                    ("z_max", max),
                ):
                    if (
                        prop not in profile["volume"]["custom_box"]
                        or profile["volume"]["custom_box"][prop] is None
                    ):
                        profile["volume"]["custom_box"][prop] = default_box[prop]
                    else:
                        value = profile["volume"]["custom_box"][prop]
                        try:
                            value = limiter(float(value), default_box[prop])
                            profile["volume"]["custom_box"][prop] = value
                        except Exception:
                            self._logger.warning(
                                "Profile has invalid value in volume.custom_box.{}: {!r}".format(
                                    prop, value
                                )
                            )
                            return False

                # make sure we actually do have a custom box and not just the same values as the
                # default box
                for prop in profile["volume"]["custom_box"]:
                    if profile["volume"]["custom_box"][prop] != default_box[prop]:
                        break
                else:
                    # exactly the same as the default box, remove custom box
                    profile["volume"]["custom_box"] = False

        # validate offsets
        offsets = []
        for offset in profile["extruder"]["offsets"]:
            if not len(offset) == 2:
                self._logger.warning(
                    "Profile has an invalid extruder.offsets entry: {entry!r}".format(
                        entry=offset
                    )
                )
                return False
            x_offset, y_offset = offset
            try:
                offsets.append((float(x_offset), float(y_offset)))
            except Exception:
                self._logger.warning(
                    "Profile has an extruder.offsets entry with non-float values: {entry!r}".format(
                        entry=offset
                    )
                )
                return False
        profile["extruder"]["offsets"] = offsets

        return profile

    @staticmethod
    def _default_box_for_volume(volume):
        if volume["origin"] == BedOrigin.CENTER:
            half_width = volume["width"] / 2
            half_depth = volume["depth"] / 2
            return {
                "x_min": -half_width,
                "x_max": half_width,
                "y_min": -half_depth,
                "y_max": half_depth,
                "z_min": 0.0,
                "z_max": volume["height"],
            }
        else:
            return {
                "x_min": 0.0,
                "x_max": volume["width"],
                "y_min": 0.0,
                "y_max": volume["depth"],
                "z_min": 0.0,
                "z_max": volume["height"],
            }



"""
This "python package" doesn't actually install. This is intentional. It is merely
used to figure out some information about the environment a specific pip call
is running under (installation dir, whether it belongs to a virtual environment,
whether the install location is writable by the current user), and for that it
only needs to be invoked by pip, the pip call doesn't have to be successful
however.

Any output (STDOUT and STDERR) produced by this script is captured by pip and,
until pip v19, printed via its STDOUT, from pip v20 on, via its STDERR. The
parsing script hence needs to capture both to support all pip versions.
"""

import os
import sys
from distutils.command.install import install as cmd_install
from distutils.dist import Distribution

cmd = cmd_install(Distribution())
cmd.finalize_options()

install_dir = cmd.install_lib
virtual_env = hasattr(sys, "real_prefix") or (
    hasattr(sys, "base_prefix") and sys.base_prefix != sys.prefix
)
writable = os.access(install_dir, os.W_OK)

lines = [
    f"PIP_INSTALL_DIR={install_dir}",
    f"PIP_VIRTUAL_ENV={virtual_env}",
    f"PIP_WRITABLE={writable}",
]

# write to stdout
for line in lines:
    print(line, file=sys.stdout)

sys.stdout.flush()

# fail intentionally
sys.exit(-1)



from typing import Dict

from pydantic import BaseModel


class StatsBase(BaseModel):
    last_version: str = ""
    """Version of OctoPrint during last start, used for keeping track of updates."""

    seen_versions: int = 0
    """Number of different versions seen."""

    server_starts: int = 0
    """Number of times OctoPrint was started."""

    prints_started: int = 0
    """Number of prints started."""

    prints_cancelled: int = 0
    """Number of prints cancelled."""

    prints_errored: int = 0
    """Number of prints errored."""

    prints_finished: int = 0
    """Number of prints finished."""

    prints_started_per_weekday: Dict[int, int] = {}
    """Number of prints started per weekday."""

    print_duration_total: float = 0
    """Total print duration."""

    print_duration_cancelled: float = 0
    """Total print duration of cancelled prints."""

    print_duration_errored: float = 0
    """Total print duration of errored prints."""

    print_duration_finished: float = 0
    """Total print duration of finished prints."""

    longest_print_duration: float = 0
    """Duration of longest print."""

    longest_print_date: int = 0
    """Timestamp of longest print."""

    files_uploaded: int = 0
    """Number of files uploaded."""

    files_deleted: int = 0
    """Number of files deleted."""

    plugins_installed: int = 0
    """Number of plugins installed."""

    plugins_uninstalled: int = 0
    """Number of plugins uninstalled."""

    most_plugins: int = 0
    """Most plugins installed at once."""


class Stats(StatsBase):
    created: int = 0
    """Timestamp of when stats collection was started."""

    created_version: str = ""
    """Version of OctoPrint when stats collection was started."""


class YearlyStats(StatsBase):
    achievements: int = 0
    """Number of achievements unlocked."""


class State(BaseModel):
    date_last_print: str = ""
    """Date of the last print."""

    prints_today: int = 0
    """Number of prints finished today."""

    date_last_cancelled_print: str = ""
    """Date of the last cancelled print."""

    prints_cancelled_today: int = 0
    """Number of prints cancelled today."""

    consecutive_prints_cancelled_today: int = 0
    """Number of consecutive prints cancelled today."""

    file_last_print: str = ""
    """Name of the file of the last print."""

    consecutive_prints_of_same_file: int = 0
    """Number of consecutive prints of the same file."""

    date_last_weekend_print: str = ""
    """Date of last print that was started on a weekend."""

    consecutive_weekend_prints: int = 0
    """Number of consecutive prints that were started on a weekend."""


class Data(BaseModel):
    stats: Stats
    achievements: Dict[str, int]
    state: State



__license__ = "GNU Affero General Public License http://www.gnu.org/licenses/agpl.html"
__copyright__ = "Copyright (C) 2015 The OctoPrint Project - Released under terms of the AGPLv3 License"


import logging
import logging.config

import click

from octoprint.cli import OctoPrintContext, get_ctx_obj_option, pass_octoprint_ctx
from octoprint.util import dict_merge

click.disable_unicode_literals_warning = True
LOGGING_CONFIG = {
    "version": 1,
    "formatters": {"brief": {"format": "%(message)s"}},
    "handlers": {
        "console": {
            "class": "logging.StreamHandler",
            "formatter": "brief",
            "stream": "ext://sys.stdout",
        }
    },
    "loggers": {"octoprint.plugin.core": {"level": logging.ERROR}},
    "root": {"level": logging.WARNING},
}

# ~~ "octoprint plugin:command" commands


class OctoPrintPluginCommands(click.MultiCommand):
    """
    Custom `click.MultiCommand <http://click.pocoo.org/5/api/#click.MultiCommand>`_
    implementation that collects commands from the plugin hook
    :ref:`octoprint.cli.commands <sec-plugins-hook-cli-commands>`.

    .. attribute:: settings

       The global :class:`~octoprint.settings.Settings` instance.

    .. attribute:: plugin_manager

       The :class:`~octoprint.plugin.core.PluginManager` instance.
    """

    sep = ":"

    def __init__(self, *args, **kwargs):
        click.MultiCommand.__init__(self, *args, **kwargs)

        self.settings = None
        self.plugin_manager = None
        self.hooks = {}

        self._logger = logging.getLogger(__name__)
        self._initialized = False

    def _initialize(self, ctx):
        if self._initialized:
            return

        click.echo("Initializing settings & plugin subsystem...")
        if ctx.obj is None:
            ctx.obj = OctoPrintContext()

        logging_config = dict_merge(
            LOGGING_CONFIG,
            {
                "root": {
                    "level": logging.DEBUG if ctx.obj.verbosity > 0 else logging.WARNING
                }
            },
        )
        logging.config.dictConfig(logging_config)

        # initialize settings and plugin manager based on provided
        # context (basedir and configfile)
        from octoprint import FatalStartupError, init_pluginsystem, init_settings

        try:
            self.settings = init_settings(
                get_ctx_obj_option(ctx, "basedir", None),
                get_ctx_obj_option(ctx, "configfile", None),
                overlays=get_ctx_obj_option(ctx, "overlays", None),
            )
            self.plugin_manager = init_pluginsystem(
                self.settings, safe_mode=get_ctx_obj_option(ctx, "safe_mode", False)
            )
        except FatalStartupError as e:
            click.echo(str(e), err=True)
            click.echo(
                "There was a fatal error initializing the settings or the plugin system.",
                err=True,
            )
            ctx.exit(-1)

        # fetch registered hooks
        self.hooks = self.plugin_manager.get_hooks("octoprint.cli.commands")

        self._initialized = True

    def list_commands(self, ctx):
        self._initialize(ctx)
        result = [name for name in self._get_commands()]
        result.sort()
        return result

    def get_command(self, ctx, cmd_name):
        self._initialize(ctx)
        commands = self._get_commands()
        return commands.get(cmd_name, None)

    def _get_commands(self):
        """Fetch all commands from plugins providing any."""

        import collections

        result = collections.OrderedDict()

        for name, hook in self.hooks.items():
            try:
                commands = hook(self, pass_octoprint_ctx)
                for command in commands:
                    if not isinstance(command, click.Command):
                        self._logger.warning(
                            "Plugin {} provided invalid CLI command, ignoring it: {!r}".format(
                                name, command
                            )
                        )
                        continue
                    result[name + self.sep + command.name] = command
            except Exception:
                self._logger.exception(
                    f"Error while retrieving cli commands for plugin {name}",
                    extra={"plugin": name},
                )

        return result


@click.group(cls=OctoPrintPluginCommands)
@pass_octoprint_ctx
def cli(obj):
    """Additional commands provided by plugins."""
    pass



__license__ = "GNU Affero General Public License http://www.gnu.org/licenses/agpl.html"
__copyright__ = "Copyright (C) 2020 The OctoPrint Project - Released under terms of the AGPLv3 License"

import re

from emoji import demojize

from octoprint.util import to_unicode
from octoprint.vendor.awesome_slugify import Slugify

_UNICODE_VARIATIONS = re.compile("[\uFE00-\uFE0F]", re.U)
_SLUGIFIES = {}


def demojify(text):
    text = to_unicode(text)
    text = remove_unicode_variations(text)
    return demojize(text, delimiters=("", ""))


def sanitize(text, safe_chars="-_.", demoji=True):
    """
    Sanitizes text by running it through slugify and optionally emoji translating.

    Examples:

    >>> sanitize("Hello World!")
    'Hello-World'
    >>> sanitize("Hello World!", safe_chars="-_. ")
    'Hello World'
    >>> sanitize("\u2764")
    'red_heart'
    >>> sanitize("\u2764\ufe00")
    'red_heart'
    >>> sanitize("\u2764", demoji=False)
    ''

    Args:
        text: the text to sanitize
        safe_chars: characters to consider safe and to keep after sanitization
        emoji: whether to also convert emoji to text

    Returns: the sanitized text
    """
    slugify = _SLUGIFIES.get(safe_chars)
    if slugify is None:
        slugify = Slugify()
        slugify.safe_chars = safe_chars
        _SLUGIFIES[safe_chars] = slugify

    text = to_unicode(text)
    if demoji:
        text = demojify(text)
    return slugify(text)


def remove_unicode_variations(text):
    return _UNICODE_VARIATIONS.sub("", text)



__license__ = "GNU Affero General Public License http://www.gnu.org/licenses/agpl.html"
__copyright__ = "Copyright (C) 2020 The OctoPrint Project - Released under terms of the AGPLv3 License"

import ast
import logging
import os
import shutil
import tempfile

from octoprint.settings import settings
from octoprint.util.net import download_file

from .. import exceptions

logger = logging.getLogger(
    "octoprint.plugins.softwareupdate.updaters.download_single_file_plugin"
)


def can_perform_update(target, check, online=True):
    return online or check.get("offline", False)


def perform_update(target, check, target_version, log_cb=None, online=True, force=False):
    if not online and not check.get("offline", False):
        raise exceptions.CannotUpdateOffline()

    url = check.get("url")
    if url is None:
        raise exceptions.ConfigurationInvalid(
            "download_single_file_plugin updater needs url set"
        )

    def _log_call(*lines):
        _log(lines, prefix=" ", stream="call")

    def _log_message(*lines):
        _log(lines, prefix="#", stream="message")

    def _log(lines, prefix=None, stream=None):
        if log_cb is None:
            return
        log_cb(lines, prefix=prefix, stream=stream)

    folder = None
    try:
        try:
            _log_message(f"Download file from {url}")
            folder = tempfile.TemporaryDirectory()
            path = download_file(url, folder.name)
        except Exception as exc:
            raise exceptions.NetworkError(cause=exc)

        filename = os.path.basename(path)
        _, ext = os.path.splitext(filename)
        if ext not in (".py",):
            raise exceptions.UpdateError(f"File is not a python file: {filename}", None)

        try:
            with open(path, "rb") as f:
                ast.parse(f.read(), filename=path)
        except Exception:
            logger.exception(f"Could not parse {path} as python file", None)
            raise exceptions.UpdateError(
                f"Could not parse {filename} as python file.", None
            )

        destination = os.path.join(settings().getBaseFolder("plugins"), filename)

        try:
            _log_message(f"Copy {path} to {destination}")
            shutil.copy(path, destination)
        except Exception:
            logger.exception(f"Could not copy {path} to {destination}")
            raise exceptions.UpdateError(f"Could not copy {path} to {destination}", None)

        return "ok"
    finally:
        if folder is not None:
            folder.cleanup()



__license__ = "GNU Affero General Public License http://www.gnu.org/licenses/agpl.html"
__copyright__ = "Copyright (C) 2022 The OctoPrint Project - Released under terms of the AGPLv3 License"

from enum import Enum

from octoprint.schema import BaseModel
from octoprint.vendor.with_attrs_docs import with_attrs_docs


class RunAtEnum(str, Enum):
    never = "never"
    idle = "idle"
    always = "always"


@with_attrs_docs
class GcodeAnalysisConfig(BaseModel):
    maxExtruders: int = 10
    """Maximum number of extruders to support/to sanity check for."""

    throttle_normalprio: float = 0.01
    """Pause between each processed GCODE line batch in normal priority mode, seconds."""

    throttle_highprio: float = 0.0
    """Pause between each processed GCODE line batch in high priority mode (e.g. on fresh uploads), seconds."""

    throttle_lines: int = 100
    """GCODE line batch size."""

    runAt: RunAtEnum = RunAtEnum.idle
    """Whether to run the analysis only when idle (not printing), regardless of printing state or never."""

    bedZ: float = 0.0
    """Z position considered the location of the bed."""



__author__ = "Gina Häußge <osd@foosel.net>"
__license__ = "GNU Affero General Public License http://www.gnu.org/licenses/agpl.html"
__copyright__ = "Copyright (C) 2014 The OctoPrint Project - Released under terms of the AGPLv3 License"

import copy
import logging
import os
import shutil
from contextlib import contextmanager
from os import scandir, walk

import pylru

import octoprint.filemanager
from octoprint.util import (
    atomic_write,
    is_hidden_path,
    time_this,
    to_bytes,
    to_unicode,
    yaml,
)
from octoprint.util.files import sanitize_filename


class StorageInterface:
    """
    Interface of storage adapters for OctoPrint.
    """

    # noinspection PyUnreachableCode
    @property
    def analysis_backlog(self):
        """
        Get an iterator over all items stored in the storage that need to be analysed by the :class:`~octoprint.filemanager.AnalysisQueue`.

        The yielded elements are expected as storage specific absolute paths to the respective files. Don't forget
        to recurse into folders if your storage adapter supports those.

        :return: an iterator yielding all un-analysed files in the storage
        """
        # empty generator pattern, yield is intentionally unreachable
        return
        yield

    # noinspection PyUnreachableCode
    def analysis_backlog_for_path(self, path=None):
        # empty generator pattern, yield is intentionally unreachable
        return
        yield

    def last_modified(self, path=None, recursive=False):
        """
        Get the last modification date of the specified ``path`` or ``path``'s subtree.

        Args:
            path (str or None): Path for which to determine the subtree's last modification date. If left out or
                set to None, defatuls to storage root.
            recursive (bool): Whether to determine only the date of the specified ``path`` (False, default) or
                the whole ``path``'s subtree (True).

        Returns: (float) The last modification date of the indicated subtree
        """
        raise NotImplementedError()

    def get_size(self, path=None, recursive=False) -> int:
        """
        Get the size of the specified ``path`` or ``path``'s subtree.

        Args:
            path (str or None): Path for which to determine the subtree's size. If left out or
                set to None, defaults to storage root.
            recursive (bool): Whether to determine only the size of the specified ``path`` (False, default) or
                the whole ``path``'s subtree (True).
        """
        raise NotImplementedError()

    def get_lastmodified(self, path: str = None, recursive: bool = False) -> int:
        """
        Get the modification date of the specified ``path`` or ``path``'s subtree.

        Args:
            path (str or None): Path for which to determine the modification date. If left our or
                set to None, defaults to storage root.
            recursive (bool): Whether to determine only the date of the specified ``path`` (False, default) or
                the whole ``path``'s subtree (True).
        """
        raise NotImplementedError()

    def file_in_path(self, path, filepath):
        """
        Returns whether the file indicated by ``file`` is inside ``path`` or not.
        :param string path: the path to check
        :param string filepath: path to the file
        :return: ``True`` if the file is inside the path, ``False`` otherwise
        """
        return NotImplementedError()

    def file_exists(self, path):
        """
        Returns whether the file indicated by ``path`` exists or not.
        :param string path: the path to check for existence
        :return: ``True`` if the file exists, ``False`` otherwise
        """
        raise NotImplementedError()

    def folder_exists(self, path):
        """
        Returns whether the folder indicated by ``path`` exists or not.
        :param string path: the path to check for existence
        :return: ``True`` if the folder exists, ``False`` otherwise
        """
        raise NotImplementedError()

    def list_files(
        self, path=None, filter=None, recursive=True, level=0, force_refresh=False
    ):
        """
        List all files in storage starting at ``path``. If ``recursive`` is set to True (the default), also dives into
        subfolders.

        An optional filter function can be supplied which will be called with a file name and file data and which has
        to return True if the file is to be included in the result or False if not.

        The data structure of the returned result will be a dictionary mapping from file names to entry data. File nodes
        will contain their metadata here, folder nodes will contain their contained files and folders. Example::

           {
             "some_folder": {
               "name": "some_folder",
               "path": "some_folder",
               "type": "folder",
               "children": {
                 "some_sub_folder": {
                   "name": "some_sub_folder",
                   "path": "some_folder/some_sub_folder",
                   "type": "folder",
                   "typePath": ["folder"],
                   "children": { ... }
                 },
                 "some_file.gcode": {
                   "name": "some_file.gcode",
                   "path": "some_folder/some_file.gcode",
                   "type": "machinecode",
                   "typePath": ["machinecode", "gcode"],
                   "hash": "<sha1 hash>",
                   "links": [ ... ],
                   ...
                 },
                 ...
               }
             "test.gcode": {
               "name": "test.gcode",
               "path": "test.gcode",
               "type": "machinecode",
               "typePath": ["machinecode", "gcode"],
               "hash": "<sha1 hash>",
               "links": [...],
               ...
             },
             "test.stl": {
               "name": "test.stl",
               "path": "test.stl",
               "type": "model",
               "typePath": ["model", "stl"],
               "hash": "<sha1 hash>",
               "links": [...],
               ...
             },
             ...
           }

        :param string path:     base path from which to recursively list all files, optional, if not supplied listing will start
                                from root of base folder
        :param function filter: a filter that matches the files that are to be returned, may be left out in which case no
                                filtering will take place
        :param bool recursive:  will also step into sub folders for building the complete list if set to True, otherwise will only
                                do one step down into sub folders to be able to populate the ``children``.
        :return: a dictionary mapping entry names to entry data that represents the whole file list
        """
        raise NotImplementedError()

    def add_folder(self, path, ignore_existing=True, display=None):
        """
        Adds a folder as ``path``

        The ``path`` will be sanitized.

        :param string path:          the path of the new folder
        :param bool ignore_existing: if set to True, no error will be raised if the folder to be added already exists
        :param str display:          display name of the folder
        :return: the sanitized name of the new folder to be used for future references to the folder
        """
        raise NotImplementedError()

    def remove_folder(self, path, recursive=True):
        """
        Removes the folder at ``path``

        :param string path:    the path of the folder to remove
        :param bool recursive: if set to True, contained folders and files will also be removed, otherwise an error will
                               be raised if the folder is not empty (apart from any metadata files) when it's to be removed
        """
        raise NotImplementedError()

    def copy_folder(self, source, destination):
        """
        Copies the folder ``source`` to ``destination``

        :param string source: path to the source folder
        :param string destination: path to destination

        :return: the path in the storage to the copy of the folder
        """
        raise NotImplementedError()

    def move_folder(self, source, destination):
        """
        Moves the folder ``source`` to ``destination``

        :param string source: path to the source folder
        :param string destination: path to destination

        :return: the new path in the storage to the folder
        """
        raise NotImplementedError()

    def add_file(
        self,
        path,
        file_object,
        printer_profile=None,
        links=None,
        allow_overwrite=False,
        display=None,
    ):
        """
        Adds the file ``file_object`` as ``path``

        :param string path:            the file's new path, will be sanitized
        :param object file_object:     a file object that provides a ``save`` method which will be called with the destination path
                                       where the object should then store its contents
        :param object printer_profile: the printer profile associated with this file (if any)
        :param list links:             any links to add with the file
        :param bool allow_overwrite:   if set to True no error will be raised if the file already exists and the existing file
                                       and its metadata will just be silently overwritten
        :param str display:            display name of the file
        :return: the sanitized name of the file to be used for future references to it
        """
        raise NotImplementedError()

    def remove_file(self, path):
        """
        Removes the file at ``path``

        Will also take care of deleting the corresponding entries
        in the metadata and deleting all links pointing to the file.

        :param string path: path of the file to remove
        """
        raise NotImplementedError()

    def copy_file(self, source, destination):
        """
        Copies the file ``source`` to ``destination``

        :param string source: path to the source file
        :param string destination: path to destination

        :return: the path in the storage to the copy of the file
        """
        raise NotImplementedError()

    def move_file(self, source, destination):
        """
        Moves the file ``source`` to ``destination``

        :param string source: path to the source file
        :param string destination: path to destination

        :return: the new path in the storage to the file
        """
        raise NotImplementedError()

    def has_analysis(self, path):
        """
        Returns whether the file at path has been analysed yet

        :param path: virtual path to the file for which to retrieve the metadata
        """
        raise NotImplementedError()

    def get_metadata(self, path):
        """
        Retrieves the metadata for the file ``path``.

        :param path: virtual path to the file for which to retrieve the metadata
        :return: the metadata associated with the file
        """
        raise NotImplementedError()

    def add_link(self, path, rel, data):
        """
        Adds a link of relation ``rel`` to file ``path`` with the given ``data``.

        The following relation types are currently supported:

          * ``model``: adds a link to a model from which the file was created/sliced, expected additional data is the ``name``
            and optionally the ``hash`` of the file to link to. If the link can be resolved against another file on the
            current ``path``, not only will it be added to the links of ``name`` but a reverse link of type ``machinecode``
            referring to ``name`` and its hash will also be added to the linked ``model`` file
          * ``machinecode``: adds a link to a file containing machine code created from the current file (model), expected
            additional data is the ``name`` and optionally the ``hash`` of the file to link to. If the link can be resolved
            against another file on the current ``path``, not only will it be added to the links of ``name`` but a reverse
            link of type ``model`` referring to ``name`` and its hash will also be added to the linked ``model`` file.
          * ``web``: adds a location on the web associated with this file (e.g. a website where to download a model),
            expected additional data is a ``href`` attribute holding the website's URL and optionally a ``retrieved``
            attribute describing when the content was retrieved

        Note that adding ``model`` links to files identifying as models or ``machinecode`` links to files identifying
        as machine code will be refused.

        :param path: path of the file for which to add a link
        :param rel: type of relation of the link to add (currently ``model``, ``machinecode`` and ``web`` are supported)
        :param data: additional data of the link to add
        """
        raise NotImplementedError()

    def remove_link(self, path, rel, data):
        """
        Removes the link consisting of ``rel`` and ``data`` from file ``name`` on ``path``.

        :param path: path of the file from which to remove the link
        :param rel: type of relation of the link to remove (currently ``model``, ``machinecode`` and ``web`` are supported)
        :param data: additional data of the link to remove, must match existing link
        """
        raise NotImplementedError()

    def get_additional_metadata(self, path, key):
        """
        Fetches additional metadata at ``key`` from the metadata of ``path``.

        :param path: the virtual path to the file for which to fetch additional metadata
        :param key: key of metadata to fetch
        """
        raise NotImplementedError()

    def set_additional_metadata(self, path, key, data, overwrite=False, merge=False):
        """
        Adds additional metadata to the metadata of ``path``. Metadata in ``data`` will be saved under ``key``.

        If ``overwrite`` is set and ``key`` already exists in ``name``'s metadata, the current value will be overwritten.

        If ``merge`` is set and ``key`` already exists and both ``data`` and the existing data under ``key`` are dictionaries,
        the two dictionaries will be merged recursively.

        :param path: the virtual path to the file for which to add additional metadata
        :param key: key of metadata to add
        :param data: metadata to add
        :param overwrite: if True and ``key`` already exists, it will be overwritten
        :param merge: if True and ``key`` already exists and both ``data`` and the existing data are dictionaries, they
                      will be merged
        """
        raise NotImplementedError()

    def remove_additional_metadata(self, path, key):
        """
        Removes additional metadata under ``key`` for ``name`` on ``path``

        :param path: the virtual path to the file for which to remove the metadata under ``key``
        :param key: the key to remove
        """
        raise NotImplementedError()

    def canonicalize(self, path):
        """
        Canonicalizes the given ``path``. The ``path`` may consist of both folder and file name, the underlying
        implementation must separate those if necessary.

        By default, this calls :func:`~octoprint.filemanager.StorageInterface.sanitize`, which also takes care
        of stripping any invalid characters.

        Args:
                path: the path to canonicalize

        Returns:
                a 2-tuple containing the canonicalized path and file name

        """
        return self.sanitize(path)

    def sanitize(self, path):
        """
        Sanitizes the given ``path``, stripping it of all invalid characters. The ``path`` may consist of both
        folder and file name, the underlying implementation must separate those if necessary and sanitize individually.

        :param string path: the path to sanitize
        :return: a 2-tuple containing the sanitized path and file name
        """
        raise NotImplementedError()

    def sanitize_path(self, path):
        """
        Sanitizes the given folder-only ``path``, stripping it of all invalid characters.
        :param string path: the path to sanitize
        :return: the sanitized path
        """
        raise NotImplementedError()

    def sanitize_name(self, name):
        """
        Sanitizes the given file ``name``, stripping it of all invalid characters.
        :param string name: the file name to sanitize
        :return: the sanitized name
        """
        raise NotImplementedError()

    def split_path(self, path):
        """
        Split ``path`` into base directory and file name.
        :param path: the path to split
        :return: a tuple (base directory, file name)
        """
        raise NotImplementedError()

    def join_path(self, *path):
        """
        Join path elements together
        :param path: path elements to join
        :return: joined representation of the path to be usable as fully qualified path for further operations
        """
        raise NotImplementedError()

    def path_on_disk(self, path):
        """
        Retrieves the path on disk for ``path``.

        Note: if the storage is not on disk and there exists no path on disk to refer to it, this method should
        raise an :class:`io.UnsupportedOperation`

        Opposite of :func:`path_in_storage`.

        :param string path: the virtual path for which to retrieve the path on disk
        :return: the path on disk to ``path``
        """
        raise NotImplementedError()

    def path_in_storage(self, path):
        """
        Retrieves the equivalent in the storage adapter for ``path``.

        Opposite of :func:`path_on_disk`.

        :param string path: the path for which to retrieve the storage path
        :return: the path in storage to ``path``
        """
        raise NotImplementedError()


class StorageError(Exception):
    UNKNOWN = "unknown"
    INVALID_DIRECTORY = "invalid_directory"
    INVALID_FILE = "invalid_file"
    INVALID_SOURCE = "invalid_source"
    INVALID_DESTINATION = "invalid_destination"
    DOES_NOT_EXIST = "does_not_exist"
    ALREADY_EXISTS = "already_exists"
    SOURCE_EQUALS_DESTINATION = "source_equals_destination"
    NOT_EMPTY = "not_empty"

    def __init__(self, message, code=None, cause=None):
        Exception.__init__(self, message)
        self.cause = cause

        if code is None:
            code = StorageError.UNKNOWN
        self.code = code


class LocalFileStorage(StorageInterface):
    """
    The ``LocalFileStorage`` is a storage implementation which holds all files, folders and metadata on disk.

    Metadata is managed inside ``.metadata.json`` files in the respective folders, indexed by the sanitized filenames
    stored within the folder. Metadata access is managed through an LRU cache to minimize access overhead.

    This storage type implements :func:`path_on_disk`.
    """

    def __init__(self, basefolder, create=False, really_universal=False):
        """
        Initializes a ``LocalFileStorage`` instance under the given ``basefolder``, creating the necessary folder
        if necessary and ``create`` is set to ``True``.

        :param string basefolder:     the path to the folder under which to create the storage
        :param bool create:           ``True`` if the folder should be created if it doesn't exist yet, ``False`` otherwise
        :param bool really_universal: ``True`` if the file names should be forced to really universal, ``False`` otherwise
        """
        self._logger = logging.getLogger(__name__)

        self.basefolder = os.path.realpath(os.path.abspath(to_unicode(basefolder)))
        if not os.path.exists(self.basefolder) and create:
            os.makedirs(self.basefolder)
        if not os.path.exists(self.basefolder) or not os.path.isdir(self.basefolder):
            raise StorageError(
                f"{basefolder} is not a valid directory",
                code=StorageError.INVALID_DIRECTORY,
            )

        self._really_universal = really_universal

        import threading

        self._metadata_lock_mutex = threading.RLock()
        self._metadata_locks = {}
        self._persisted_metadata_lock_mutex = threading.RLock()
        self._persisted_metadata_locks = {}

        self._metadata_cache = pylru.lrucache(100)
        self._filelist_cache = {}
        self._filelist_cache_mutex = threading.RLock()

        self._old_metadata = None
        self._initialize_metadata()

    def _initialize_metadata(self):
        self._logger.info(f"Initializing the file metadata for {self.basefolder}...")

        old_metadata_path = os.path.join(self.basefolder, "metadata.yaml")
        backup_path = os.path.join(self.basefolder, "metadata.yaml.backup")

        if os.path.exists(old_metadata_path):
            # load the old metadata file
            try:
                self._old_metadata = yaml.load_from_file(path=old_metadata_path)
            except Exception:
                self._logger.exception("Error while loading old metadata file")

            # make sure the metadata is initialized as far as possible
            self._list_folder(self.basefolder)

            # rename the old metadata file
            self._old_metadata = None
            try:
                import shutil

                shutil.move(old_metadata_path, backup_path)
            except Exception:
                self._logger.exception("Could not rename old metadata.yaml file")

        else:
            # make sure the metadata is initialized as far as possible
            self._list_folder(self.basefolder)

        self._logger.info(
            f"... file metadata for {self.basefolder} initialized successfully."
        )

    @property
    def analysis_backlog(self):
        return self.analysis_backlog_for_path()

    def analysis_backlog_for_path(self, path=None):
        if path:
            path = self.sanitize_path(path)

        yield from self._analysis_backlog_generator(path)

    def _analysis_backlog_generator(self, path=None):
        if path is None:
            path = self.basefolder

        metadata = self._get_metadata(path)
        if not metadata:
            metadata = {}
        for entry in scandir(path):
            if is_hidden_path(entry.name):
                continue

            if entry.is_file() and octoprint.filemanager.valid_file_type(entry.name):
                if (
                    entry.name not in metadata
                    or not isinstance(metadata[entry.name], dict)
                    or "analysis" not in metadata[entry.name]
                ):
                    printer_profile_rels = self.get_link(entry.path, "printerprofile")
                    if printer_profile_rels:
                        printer_profile_id = printer_profile_rels[0]["id"]
                    else:
                        printer_profile_id = None

                    yield entry.name, entry.path, printer_profile_id
            elif os.path.isdir(entry.path):
                for sub_entry in self._analysis_backlog_generator(entry.path):
                    yield (
                        self.join_path(entry.name, sub_entry[0]),
                        sub_entry[1],
                        sub_entry[2],
                    )

    def last_modified(self, path=None, recursive=False):
        if path is None:
            path = self.basefolder
        else:
            path = os.path.join(self.basefolder, path)

        def last_modified_for_path(p):
            metadata = os.path.join(p, ".metadata.json")
            if os.path.exists(metadata):
                return max(os.stat(p).st_mtime, os.stat(metadata).st_mtime)
            else:
                return os.stat(p).st_mtime

        if recursive:
            return max(last_modified_for_path(root) for root, _, _ in walk(path))
        else:
            return last_modified_for_path(path)

    def get_size(self, path=None, recursive=False):
        if path is None:
            path = self.basefolder

        path, name = self.sanitize(path)
        path = os.path.join(path, name)

        # shortcut for individual files
        if os.path.isfile(path):
            return os.stat(path).st_size

        size = 0
        for entry in os.scandir(path):
            if entry.is_file():
                size += entry.stat().st_size
            elif recursive and entry.is_dir():
                size += self.get_size(entry.path, recursive=recursive)

        return size

    def get_lastmodified(self, path: str = None, recursive: bool = False) -> int:
        if path is None:
            path = self.basefolder

        path, name = self.sanitize(path)
        path = os.path.join(path, name)

        # shortcut for individual files
        if os.path.isfile(path):
            return int(os.stat(path).st_mtime)

        last_modified = 0
        for entry in os.scandir(path):
            if entry.is_file():
                last_modified = max(last_modified, entry.stat().st_mtime)
            elif recursive and entry.is_dir():
                last_modified = max(
                    last_modified,
                    self.get_lastmodified(entry.path, recursive=recursive),
                )

        return int(last_modified)

    def file_in_path(self, path, filepath):
        filepath = self.sanitize_path(filepath)
        path = self.sanitize_path(path)

        return filepath == path or filepath.startswith(path + os.sep)

    def file_exists(self, path):
        path, name = self.sanitize(path)
        file_path = os.path.join(path, name)
        return os.path.exists(file_path) and os.path.isfile(file_path)

    def folder_exists(self, path):
        path, name = self.sanitize(path)
        folder_path = os.path.join(path, name)
        return os.path.exists(folder_path) and os.path.isdir(folder_path)

    def list_files(
        self, path=None, filter=None, recursive=True, level=0, force_refresh=False
    ):
        if path:
            path = self.sanitize_path(to_unicode(path))
            base = self.path_in_storage(path)
            if base:
                base += "/"
        else:
            path = self.basefolder
            base = ""

        def strip_children(nodes):
            result = {}
            for key, node in nodes.items():
                if node["type"] == "folder":
                    node = copy.copy(node)
                    node["children"] = {}
                result[key] = node
            return result

        def strip_grandchildren(nodes):
            result = {}
            for key, node in nodes.items():
                if node["type"] == "folder":
                    node = copy.copy(node)
                    node["children"] = strip_children(node["children"])
                result[key] = node
            return result

        def apply_filter(nodes, filter_func):
            result = {}
            for key, node in nodes.items():
                if filter_func(node) or node["type"] == "folder":
                    if node["type"] == "folder":
                        node = copy.copy(node)
                        node["children"] = apply_filter(
                            node.get("children", {}), filter_func
                        )
                    result[key] = node
            return result

        result = self._list_folder(path, base=base, force_refresh=force_refresh)
        if not recursive:
            if level > 0:
                result = strip_grandchildren(result)
            else:
                result = strip_children(result)
        if callable(filter):
            result = apply_filter(result, filter)
        return result

    def add_folder(self, path, ignore_existing=True, display=None):
        display_path, display_name = self.canonicalize(path)
        path = self.sanitize_path(display_path)
        name = self.sanitize_name(display_name)

        if display is not None:
            display_name = display

        folder_path = os.path.join(path, name)
        if os.path.exists(folder_path):
            if not ignore_existing:
                raise StorageError(
                    f"{name} does already exist in {path}",
                    code=StorageError.ALREADY_EXISTS,
                )
        else:
            os.mkdir(folder_path)

        if display_name != name:
            metadata = self._get_metadata_entry(path, name, default={})
            metadata["display"] = display_name
            self._update_metadata_entry(path, name, metadata)

        return self.path_in_storage((path, name))

    def remove_folder(self, path, recursive=True):
        path, name = self.sanitize(path)

        folder_path = os.path.join(path, name)
        if not os.path.exists(folder_path):
            return

        empty = True
        for entry in scandir(folder_path):
            if entry.name == ".metadata.json" or entry.name == ".metadata.yaml":
                continue
            empty = False
            break

        if not empty and not recursive:
            raise StorageError(
                f"{name} in {path} is not empty",
                code=StorageError.NOT_EMPTY,
            )

        import shutil

        shutil.rmtree(folder_path)

        self._remove_metadata_entry(path, name)

    def _get_source_destination_data(self, source, destination, must_not_equal=False):
        """Prepares data dicts about source and destination for copy/move."""
        source_path, source_name = self.sanitize(source)

        destination_canon_path, destination_canon_name = self.canonicalize(destination)
        destination_path = self.sanitize_path(destination_canon_path)
        destination_name = self.sanitize_name(destination_canon_name)

        source_fullpath = os.path.join(source_path, source_name)
        destination_fullpath = os.path.join(destination_path, destination_name)

        if not os.path.exists(source_fullpath):
            raise StorageError(
                f"{source_name} in {source_path} does not exist",
                code=StorageError.INVALID_SOURCE,
            )

        if not os.path.isdir(destination_path):
            raise StorageError(
                "Destination path {} does not exist or is not a folder".format(
                    destination_path
                ),
                code=StorageError.INVALID_DESTINATION,
            )
        if (
            os.path.exists(destination_fullpath)
            and source_fullpath != destination_fullpath
        ):
            raise StorageError(
                f"{destination_name} does already exist in {destination_path}",
                code=StorageError.ALREADY_EXISTS,
            )

        source_meta = self._get_metadata_entry(source_path, source_name)
        if source_meta:
            source_display = source_meta.get("display", source_name)
        else:
            source_display = source_name

        if (
            must_not_equal or source_display == destination_canon_name
        ) and source_fullpath == destination_fullpath:
            raise StorageError(
                "Source {} and destination {} are the same folder".format(
                    source_path, destination_path
                ),
                code=StorageError.SOURCE_EQUALS_DESTINATION,
            )

        source_data = {
            "path": source_path,
            "name": source_name,
            "display": source_display,
            "fullpath": source_fullpath,
        }
        destination_data = {
            "path": destination_path,
            "name": destination_name,
            "display": destination_canon_name,
            "fullpath": destination_fullpath,
        }
        return source_data, destination_data

    def _set_display_metadata(self, destination_data, source_data=None):
        if (
            source_data
            and destination_data["name"] == source_data["name"]
            and source_data["name"] != source_data["display"]
        ):
            display = source_data["display"]
        elif destination_data["name"] != destination_data["display"]:
            display = destination_data["display"]
        else:
            display = None

        destination_meta = self._get_metadata_entry(
            destination_data["path"], destination_data["name"], default={}
        )
        if display:
            destination_meta["display"] = display
            self._update_metadata_entry(
                destination_data["path"], destination_data["name"], destination_meta
            )
        elif "display" in destination_meta:
            del destination_meta["display"]
            self._update_metadata_entry(
                destination_data["path"], destination_data["name"], destination_meta
            )

    def copy_folder(self, source, destination):
        source_data, destination_data = self._get_source_destination_data(
            source, destination, must_not_equal=True
        )

        try:
            shutil.copytree(source_data["fullpath"], destination_data["fullpath"])
        except Exception as e:
            raise StorageError(
                "Could not copy %s in %s to %s in %s"
                % (
                    source_data["name"],
                    source_data["path"],
                    destination_data["name"],
                    destination_data["path"],
                ),
                cause=e,
            )

        self._set_display_metadata(destination_data, source_data=source_data)

        return self.path_in_storage(destination_data["fullpath"])

    def move_folder(self, source, destination):
        source_data, destination_data = self._get_source_destination_data(
            source, destination
        )

        # only a display rename? Update that and bail early
        if source_data["fullpath"] == destination_data["fullpath"]:
            self._set_display_metadata(destination_data)
            return self.path_in_storage(destination_data["fullpath"])

        try:
            shutil.move(source_data["fullpath"], destination_data["fullpath"])
        except Exception as e:
            raise StorageError(
                "Could not move %s in %s to %s in %s"
                % (
                    source_data["name"],
                    source_data["path"],
                    destination_data["name"],
                    destination_data["path"],
                ),
                cause=e,
            )

        self._set_display_metadata(destination_data, source_data=source_data)
        self._remove_metadata_entry(source_data["path"], source_data["name"])
        self._delete_metadata(source_data["fullpath"])

        return self.path_in_storage(destination_data["fullpath"])

    def add_file(
        self,
        path,
        file_object,
        printer_profile=None,
        links=None,
        allow_overwrite=False,
        display=None,
    ):
        display_path, display_name = self.canonicalize(path)
        path = self.sanitize_path(display_path)
        name = self.sanitize_name(display_name)

        if display:
            display_name = display

        if not octoprint.filemanager.valid_file_type(name):
            raise StorageError(
                f"{name} is an unrecognized file type",
                code=StorageError.INVALID_FILE,
            )

        file_path = os.path.join(path, name)
        if os.path.exists(file_path) and not os.path.isfile(file_path):
            raise StorageError(
                f"{name} does already exist in {path} and is not a file",
                code=StorageError.ALREADY_EXISTS,
            )
        if os.path.exists(file_path) and not allow_overwrite:
            raise StorageError(
                f"{name} does already exist in {path} and overwriting is prohibited",
                code=StorageError.ALREADY_EXISTS,
            )

        # make sure folders exist
        if not os.path.exists(path):
            # TODO persist display names of path segments!
            os.makedirs(path)

        # save the file
        file_object.save(file_path)

        # save the file's hash to the metadata of the folder
        file_hash = self._create_hash(file_path)
        metadata = self._get_metadata_entry(path, name, default={})
        metadata_dirty = False
        if "hash" not in metadata or metadata["hash"] != file_hash:
            # hash changed -> throw away old metadata
            metadata = {"hash": file_hash}
            metadata_dirty = True

        if "display" not in metadata and display_name != name:
            # display name is not the same as file name -> store in metadata
            metadata["display"] = display_name
            metadata_dirty = True

        if metadata_dirty:
            self._update_metadata_entry(path, name, metadata)

        # process any links that were also provided for adding to the file
        if not links:
            links = []

        if printer_profile is not None:
            links.append(
                (
                    "printerprofile",
                    {"id": printer_profile["id"], "name": printer_profile["name"]},
                )
            )

        self._add_links(name, path, links)

        # touch the file to set last access and modification time to now
        os.utime(file_path, None)

        return self.path_in_storage((path, name))

    def remove_file(self, path):
        path, name = self.sanitize(path)

        file_path = os.path.join(path, name)
        if not os.path.exists(file_path):
            return
        if not os.path.isfile(file_path):
            raise StorageError(
                f"{name} in {path} is not a file",
                code=StorageError.INVALID_FILE,
            )

        try:
            os.remove(file_path)
        except Exception as e:
            raise StorageError(f"Could not delete {name} in {path}", cause=e)

        self._remove_metadata_entry(path, name)

    def copy_file(self, source, destination):
        source_data, destination_data = self._get_source_destination_data(
            source, destination, must_not_equal=True
        )

        if not octoprint.filemanager.valid_file_type(destination_data["name"]):
            raise StorageError(
                f"{destination_data['name']} is an unrecognized file type",
                code=StorageError.INVALID_FILE,
            )

        try:
            shutil.copy2(source_data["fullpath"], destination_data["fullpath"])
        except Exception as e:
            raise StorageError(
                "Could not copy %s in %s to %s in %s"
                % (
                    source_data["name"],
                    source_data["path"],
                    destination_data["name"],
                    destination_data["path"],
                ),
                cause=e,
            )

        self._copy_metadata_entry(
            source_data["path"],
            source_data["name"],
            destination_data["path"],
            destination_data["name"],
        )
        self._set_display_metadata(destination_data, source_data=source_data)

        return self.path_in_storage(destination_data["fullpath"])

    def move_file(self, source, destination, allow_overwrite=False):
        source_data, destination_data = self._get_source_destination_data(
            source, destination
        )

        if not octoprint.filemanager.valid_file_type(destination_data["name"]):
            raise StorageError(
                f"{destination_data['name']} is an unrecognized file type",
                code=StorageError.INVALID_FILE,
            )

        # only a display rename? Update that and bail early
        if source_data["fullpath"] == destination_data["fullpath"]:
            self._set_display_metadata(destination_data)
            return self.path_in_storage(destination_data["fullpath"])

        try:
            shutil.move(source_data["fullpath"], destination_data["fullpath"])
        except Exception as e:
            raise StorageError(
                "Could not move %s in %s to %s in %s"
                % (
                    source_data["name"],
                    source_data["path"],
                    destination_data["name"],
                    destination_data["path"],
                ),
                cause=e,
            )

        self._copy_metadata_entry(
            source_data["path"],
            source_data["name"],
            destination_data["path"],
            destination_data["name"],
            delete_source=True,
        )
        self._set_display_metadata(destination_data, source_data=source_data)

        return self.path_in_storage(destination_data["fullpath"])

    def has_analysis(self, path):
        metadata = self.get_metadata(path)
        return "analysis" in metadata

    def get_metadata(self, path):
        path, name = self.sanitize(path)
        return self._get_metadata_entry(path, name)

    def get_link(self, path, rel):
        path, name = self.sanitize(path)
        return self._get_links(name, path, rel)

    def add_link(self, path, rel, data):
        path, name = self.sanitize(path)
        self._add_links(name, path, [(rel, data)])

    def remove_link(self, path, rel, data):
        path, name = self.sanitize(path)
        self._remove_links(name, path, [(rel, data)])

    def add_history(self, path, data):
        path, name = self.sanitize(path)
        self._add_history(name, path, data)

    def update_history(self, path, index, data):
        path, name = self.sanitize(path)
        self._update_history(name, path, index, data)

    def remove_history(self, path, index):
        path, name = self.sanitize(path)
        self._delete_history(name, path, index)

    def get_additional_metadata(self, path, key):
        path, name = self.sanitize(path)
        metadata = self._get_metadata(path)

        if name not in metadata:
            return

        return metadata[name].get(key)

    def set_additional_metadata(self, path, key, data, overwrite=False, merge=False):
        path, name = self.sanitize(path)
        metadata = self._get_metadata(path)
        metadata_dirty = False

        if name not in metadata:
            return

        metadata = self._copied_metadata(metadata, name)

        if key not in metadata[name] or overwrite:
            metadata[name][key] = data
            metadata_dirty = True
        elif (
            key in metadata[name]
            and isinstance(metadata[name][key], dict)
            and isinstance(data, dict)
            and merge
        ):
            import octoprint.util

            metadata[name][key] = octoprint.util.dict_merge(
                metadata[name][key], data, in_place=True
            )
            metadata_dirty = True

        if metadata_dirty:
            self._save_metadata(path, metadata)

    def remove_additional_metadata(self, path, key):
        path, name = self.sanitize(path)
        metadata = self._get_metadata(path)

        if name not in metadata:
            return

        if key not in metadata[name]:
            return

        metadata = self._copied_metadata(metadata, name)
        del metadata[name][key]
        self._save_metadata(path, metadata)

    def split_path(self, path):
        path = to_unicode(path)
        split = path.split("/")

        if len(split) == 1:
            return "", split[0]

        return self.path_in_storage(self.join_path(*split[:-1])), split[-1]

    def join_path(self, *path):
        return self.path_in_storage("/".join(map(to_unicode, path)))

    def sanitize(self, path):
        """
        Returns a ``(path, name)`` tuple derived from the provided ``path``.

        ``path`` may be:
          * a storage path
          * an absolute file system path
          * a tuple or list containing all individual path elements
          * a string representation of the path
          * with or without a file name

        Note that for a ``path`` without a trailing slash the last part will be considered a file name and
        hence be returned at second position. If you only need to convert a folder path, be sure to
        include a trailing slash for a string ``path`` or an empty last element for a list ``path``.
        """

        path, name = self.canonicalize(path)
        name = self.sanitize_name(name)
        path = self.sanitize_path(path)
        return path, name

    def canonicalize(self, path):
        name = None
        if isinstance(path, str):
            path = to_unicode(path)
            if path.startswith(self.basefolder):
                path = path[len(self.basefolder) :]
            path = path.replace(os.path.sep, "/")
            path = path.split("/")
        if isinstance(path, (list, tuple)):
            if len(path) == 1:
                name = to_unicode(path[0])
                path = ""
            else:
                name = to_unicode(path[-1])
                path = self.join_path(*map(to_unicode, path[:-1]))
        if not path:
            path = ""

        return path, name

    def sanitize_name(self, name):
        """
        Raises a :class:`ValueError` for a ``name`` containing ``/`` or ``\\``. Otherwise
        sanitizes the given ``name`` using ``octoprint.files.sanitize_filename``. Also
        strips any leading ``.``.
        """
        return sanitize_filename(name, really_universal=self._really_universal)

    def sanitize_path(self, path):
        """
        Ensures that the on disk representation of ``path`` is located under the configured basefolder. Resolves all
        relative path elements (e.g. ``..``) and sanitizes folder names using :func:`sanitize_name`. Final path is the
        absolute path including leading ``basefolder`` path.
        """
        path = to_unicode(path)

        if len(path):
            if path[0] == "/":
                path = path[1:]
            elif path[0] == "." and path[1] == "/":
                path = path[2:]

        path_elements = path.split("/")
        joined_path = self.basefolder
        for path_element in path_elements:
            if path_element == ".." or path_element == ".":
                joined_path = os.path.join(joined_path, path_element)
            else:
                joined_path = os.path.join(joined_path, self.sanitize_name(path_element))
        path = os.path.realpath(joined_path)
        if not path.startswith(self.basefolder):
            raise ValueError(f"path not contained in base folder: {path}")
        return path

    def _sanitize_entry(self, entry, path, entry_path):
        entry = to_unicode(entry)
        sanitized = self.sanitize_name(entry)
        if sanitized != entry:
            # entry is not sanitized yet, let's take care of that
            sanitized_path = os.path.join(path, sanitized)
            sanitized_name, sanitized_ext = os.path.splitext(sanitized)

            counter = 1
            while os.path.exists(sanitized_path):
                counter += 1
                sanitized = self.sanitize_name(
                    f"{sanitized_name}_({counter}){sanitized_ext}"
                )
                sanitized_path = os.path.join(path, sanitized)

            try:
                shutil.move(entry_path, sanitized_path)

                self._logger.info(f'Sanitized "{entry_path}" to "{sanitized_path}"')
                return sanitized, sanitized_path
            except Exception:
                self._logger.exception(
                    'Error while trying to rename "{}" to "{}", ignoring file'.format(
                        entry_path, sanitized_path
                    )
                )
                raise

        return entry, entry_path

    def path_in_storage(self, path):
        if isinstance(path, (tuple, list)):
            path = self.join_path(*path)
        if isinstance(path, str):
            path = to_unicode(path)
            if path.startswith(self.basefolder):
                path = path[len(self.basefolder) :]
            path = path.replace(os.path.sep, "/")
        while path.startswith("/"):
            path = path[1:]

        return path

    def path_on_disk(self, path):
        path, name = self.sanitize(path)
        return os.path.join(path, name)

    ##~~ internals

    def _add_history(self, name, path, data):
        metadata = self._copied_metadata(self._get_metadata(path), name)

        if "hash" not in metadata[name]:
            metadata[name]["hash"] = self._create_hash(os.path.join(path, name))

        if "history" not in metadata[name]:
            metadata[name]["history"] = []

        metadata[name]["history"].append(data)
        self._calculate_stats_from_history(name, path, metadata=metadata, save=False)
        self._save_metadata(path, metadata)

    def _update_history(self, name, path, index, data):
        metadata = self._get_metadata(path)

        if name not in metadata or "history" not in metadata[name]:
            return

        metadata = self._copied_metadata(metadata, name)

        try:
            metadata[name]["history"][index].update(data)
            self._calculate_stats_from_history(name, path, metadata=metadata, save=False)
            self._save_metadata(path, metadata)
        except IndexError:
            pass

    def _delete_history(self, name, path, index):
        metadata = self._get_metadata(path)

        if name not in metadata or "history" not in metadata[name]:
            return

        metadata = self._copied_metadata(metadata, name)

        try:
            del metadata[name]["history"][index]
            self._calculate_stats_from_history(name, path, metadata=metadata, save=False)
            self._save_metadata(path, metadata)
        except IndexError:
            pass

    def _calculate_stats_from_history(self, name, path, metadata=None, save=True):
        if metadata is None:
            metadata = self._copied_metadata(self._get_metadata(path), name)

        if "history" not in metadata[name]:
            return

        # collect data from history
        former_print_times = {}
        last_print = {}

        for history_entry in metadata[name]["history"]:
            if (
                "printTime" not in history_entry
                or "success" not in history_entry
                or not history_entry["success"]
                or "printerProfile" not in history_entry
            ):
                continue

            printer_profile = history_entry["printerProfile"]
            if not printer_profile:
                continue

            print_time = history_entry["printTime"]
            try:
                print_time = float(print_time)
            except Exception:
                self._logger.warning(
                    "Invalid print time value found in print history for {} in {}/.metadata.json: {!r}".format(
                        name, path, print_time
                    )
                )
                continue

            if printer_profile not in former_print_times:
                former_print_times[printer_profile] = []
            former_print_times[printer_profile].append(print_time)

            if (
                printer_profile not in last_print
                or last_print[printer_profile] is None
                or (
                    "timestamp" in history_entry
                    and history_entry["timestamp"]
                    > last_print[printer_profile]["timestamp"]
                )
            ):
                last_print[printer_profile] = history_entry

        # calculate stats
        statistics = {"averagePrintTime": {}, "lastPrintTime": {}}

        for printer_profile in former_print_times:
            if not former_print_times[printer_profile]:
                continue
            statistics["averagePrintTime"][printer_profile] = sum(
                former_print_times[printer_profile]
            ) / len(former_print_times[printer_profile])

        for printer_profile in last_print:
            if not last_print[printer_profile]:
                continue
            statistics["lastPrintTime"][printer_profile] = last_print[printer_profile][
                "printTime"
            ]

        metadata[name]["statistics"] = statistics

        if save:
            self._save_metadata(path, metadata)

    def _get_links(self, name, path, searched_rel):
        metadata = self._get_metadata(path)
        result = []

        if name not in metadata:
            return result

        if "links" not in metadata[name]:
            return result

        for data in metadata[name]["links"]:
            if "rel" not in data or not data["rel"] == searched_rel:
                continue
            result.append(data)
        return result

    def _add_links(self, name, path, links):
        file_type = octoprint.filemanager.get_file_type(name)
        if file_type:
            file_type = file_type[0]

        metadata = self._copied_metadata(self._get_metadata(path), name)
        metadata_dirty = False

        if "hash" not in metadata[name]:
            metadata[name]["hash"] = self._create_hash(os.path.join(path, name))

        if "links" not in metadata[name]:
            metadata[name]["links"] = []

        for rel, data in links:
            if (rel == "model" or rel == "machinecode") and "name" in data:
                if file_type == "model" and rel == "model":
                    # adding a model link to a model doesn't make sense
                    return
                elif file_type == "machinecode" and rel == "machinecode":
                    # adding a machinecode link to a machinecode doesn't make sense
                    return

                ref_path = os.path.join(path, data["name"])
                if not os.path.exists(ref_path):
                    # file doesn't exist, we won't create the link
                    continue

                # fetch hash of target file
                if data["name"] in metadata and "hash" in metadata[data["name"]]:
                    hash = metadata[data["name"]]["hash"]
                else:
                    hash = self._create_hash(ref_path)
                    if data["name"] not in metadata:
                        metadata[data["name"]] = {"hash": hash, "links": []}
                    else:
                        metadata[data["name"]]["hash"] = hash

                if "hash" in data and not data["hash"] == hash:
                    # file doesn't have the correct hash, we won't create the link
                    continue

                if "links" not in metadata[data["name"]]:
                    metadata[data["name"]]["links"] = []

                # add reverse link to link target file
                metadata[data["name"]]["links"].append(
                    {
                        "rel": "machinecode" if rel == "model" else "model",
                        "name": name,
                        "hash": metadata[name]["hash"],
                    }
                )
                metadata_dirty = True

                link_dict = {"rel": rel, "name": data["name"], "hash": hash}

            elif rel == "web" and "href" in data:
                link_dict = {"rel": rel, "href": data["href"]}
                if "retrieved" in data:
                    link_dict["retrieved"] = data["retrieved"]

            else:
                continue

            if link_dict:
                metadata[name]["links"].append(link_dict)
                metadata_dirty = True

        if metadata_dirty:
            self._save_metadata(path, metadata)

    def _remove_links(self, name, path, links):
        metadata = self._copied_metadata(self._get_metadata(path), name)
        metadata_dirty = False

        hash = metadata[name].get("hash", self._create_hash(os.path.join(path, name)))

        for rel, data in links:
            if (rel == "model" or rel == "machinecode") and "name" in data:
                if data["name"] in metadata and "links" in metadata[data["name"]]:
                    ref_rel = "model" if rel == "machinecode" else "machinecode"
                    for link in metadata[data["name"]]["links"]:
                        if (
                            link["rel"] == ref_rel
                            and "name" in link
                            and link["name"] == name
                            and "hash" in link
                            and link["hash"] == hash
                        ):
                            metadata[data["name"]] = copy.deepcopy(metadata[data["name"]])
                            metadata[data["name"]]["links"].remove(link)
                            metadata_dirty = True

            if "links" in metadata[name]:
                for link in metadata[name]["links"]:
                    if not link["rel"] == rel:
                        continue

                    matches = True
                    for k, v in data.items():
                        if k not in link or not link[k] == v:
                            matches = False
                            break

                    if not matches:
                        continue

                    metadata[name]["links"].remove(link)
                    metadata_dirty = True

        if metadata_dirty:
            self._save_metadata(path, metadata)

    @time_this(
        logtarget=__name__ + ".timings",
        message="{func}({func_args},{func_kwargs}) took {timing:.2f}ms",
        incl_func_args=True,
        log_enter=True,
    )
    def _list_folder(self, path, base="", force_refresh=False, **kwargs):
        def get_size(nodes):
            total_size = 0
            for node in nodes.values():
                if "size" in node:
                    total_size += node["size"]
            return total_size

        def enrich_folders(nodes):
            nodes = copy.copy(nodes)
            for key, value in nodes.items():
                if value["type"] == "folder":
                    value = copy.copy(value)
                    value["children"] = self._list_folder(
                        os.path.join(path, key),
                        base=value["path"] + "/",
                        force_refresh=force_refresh,
                    )
                    value["size"] = get_size(value["children"])
                    nodes[key] = value
            return nodes

        metadata_dirty = False
        try:
            with self._filelist_cache_mutex:
                cache = self._filelist_cache.get(path)
                lm = self.last_modified(path, recursive=True)
                if not force_refresh and cache and cache[0] >= lm:
                    return enrich_folders(cache[1])

                metadata = self._get_metadata(path)
                if not metadata:
                    metadata = {}

                result = {}

                for entry in scandir(path):
                    if is_hidden_path(entry.name):
                        # no hidden files and folders
                        continue

                    try:
                        entry_name = entry_display = entry.name
                        entry_path = entry.path
                        entry_is_file = entry.is_file()
                        entry_is_dir = entry.is_dir()
                        entry_stat = entry.stat()
                    except Exception:
                        # error while trying to fetch file metadata, that might be thanks to file already having
                        # been moved or deleted - ignore it and continue
                        continue

                    try:
                        new_entry_name, new_entry_path = self._sanitize_entry(
                            entry_name, path, entry_path
                        )
                        if entry_name != new_entry_name or entry_path != new_entry_path:
                            entry_display = to_unicode(entry_name)
                            entry_name = new_entry_name
                            entry_path = new_entry_path
                            entry_stat = os.stat(entry_path)
                    except Exception:
                        # error while trying to rename the file, we'll continue here and ignore it
                        continue

                    path_in_location = entry_name if not base else base + entry_name

                    try:
                        # file handling
                        if entry_is_file:
                            type_path = octoprint.filemanager.get_file_type(entry_name)
                            if not type_path:
                                # only supported extensions
                                continue
                            else:
                                file_type = type_path[0]

                            if entry_name in metadata and isinstance(
                                metadata[entry_name], dict
                            ):
                                entry_metadata = metadata[entry_name]
                                if (
                                    "display" not in entry_metadata
                                    and entry_display != entry_name
                                ):
                                    if not metadata_dirty:
                                        metadata = self._copied_metadata(
                                            metadata, entry_name
                                        )
                                    metadata[entry_name]["display"] = entry_display
                                    entry_metadata["display"] = entry_display
                                    metadata_dirty = True
                            else:
                                if not metadata_dirty:
                                    metadata = self._copied_metadata(metadata, entry_name)
                                entry_metadata = self._add_basic_metadata(
                                    path,
                                    entry_name,
                                    display_name=entry_display,
                                    save=False,
                                    metadata=metadata,
                                )
                                metadata_dirty = True

                            extended_entry_data = {}
                            extended_entry_data.update(entry_metadata)
                            extended_entry_data["name"] = entry_name
                            extended_entry_data["display"] = entry_metadata.get(
                                "display", entry_name
                            )
                            extended_entry_data["path"] = path_in_location
                            extended_entry_data["type"] = file_type
                            extended_entry_data["typePath"] = type_path
                            stat = entry_stat
                            if stat:
                                extended_entry_data["size"] = stat.st_size
                                extended_entry_data["date"] = int(stat.st_mtime)

                            result[entry_name] = extended_entry_data

                        # folder recursion
                        elif entry_is_dir:
                            if entry_name in metadata and isinstance(
                                metadata[entry_name], dict
                            ):
                                entry_metadata = metadata[entry_name]
                                if (
                                    "display" not in entry_metadata
                                    and entry_display != entry_name
                                ):
                                    if not metadata_dirty:
                                        metadata = self._copied_metadata(
                                            metadata, entry_name
                                        )
                                    metadata[entry_name]["display"] = entry_display
                                    entry_metadata["display"] = entry_display
                                    metadata_dirty = True
                            elif entry_name != entry_display:
                                if not metadata_dirty:
                                    metadata = self._copied_metadata(metadata, entry_name)
                                entry_metadata = self._add_basic_metadata(
                                    path,
                                    entry_name,
                                    display_name=entry_display,
                                    save=False,
                                    metadata=metadata,
                                )
                                metadata_dirty = True
                            else:
                                entry_metadata = {}

                            entry_data = {
                                "name": entry_name,
                                "display": entry_metadata.get("display", entry_name),
                                "path": path_in_location,
                                "type": "folder",
                                "typePath": ["folder"],
                            }
                            if entry_stat:
                                entry_data["date"] = int(entry_stat.st_mtime)

                            result[entry_name] = entry_data
                    except Exception:
                        # So something went wrong somewhere while processing this file entry - log that and continue
                        self._logger.exception(
                            f"Error while processing entry {entry_path}"
                        )
                        continue

                self._filelist_cache[path] = (
                    lm,
                    result,
                )
                return enrich_folders(result)
        finally:
            # save metadata
            if metadata_dirty:
                self._save_metadata(path, metadata)

    def _add_basic_metadata(
        self,
        path,
        entry,
        display_name=None,
        additional_metadata=None,
        save=True,
        metadata=None,
    ):
        if additional_metadata is None:
            additional_metadata = {}

        if metadata is None:
            metadata = self._get_metadata(path)

        entry_path = os.path.join(path, entry)

        if os.path.isfile(entry_path):
            entry_data = {
                "hash": self._create_hash(os.path.join(path, entry)),
                "links": [],
                "notes": [],
            }
            if (
                path == self.basefolder
                and self._old_metadata is not None
                and entry in self._old_metadata
                and "gcodeAnalysis" in self._old_metadata[entry]
            ):
                # if there is still old metadata available and that contains an analysis for this file, use it!
                entry_data["analysis"] = self._old_metadata[entry]["gcodeAnalysis"]

        elif os.path.isdir(entry_path):
            entry_data = {}

        else:
            return

        if display_name is not None and not display_name == entry:
            entry_data["display"] = display_name

        entry_data.update(additional_metadata)

        metadata = copy.copy(metadata)
        metadata[entry] = entry_data

        if save:
            self._save_metadata(path, metadata)

        return entry_data

    def _create_hash(self, path):
        import hashlib

        blocksize = 65536
        hash = hashlib.sha1()
        with open(path, "rb") as f:
            buffer = f.read(blocksize)
            while len(buffer) > 0:
                hash.update(buffer)
                buffer = f.read(blocksize)

        return hash.hexdigest()

    def _get_metadata_entry(self, path, name, default=None):
        with self._get_metadata_lock(path):
            metadata = self._get_metadata(path)
            return metadata.get(name, default)

    def _remove_metadata_entry(self, path, name):
        with self._get_metadata_lock(path):
            metadata = self._get_metadata(path)
            if name not in metadata:
                return

            metadata = copy.copy(metadata)

            if "hash" in metadata[name]:
                hash = metadata[name]["hash"]
                for m in metadata.values():
                    if "links" not in m:
                        continue
                    links_hash = (
                        lambda link: "hash" in link
                        and link["hash"] == hash
                        and "rel" in link
                        and (link["rel"] == "model" or link["rel"] == "machinecode")
                    )
                    m["links"] = [link for link in m["links"] if not links_hash(link)]

            del metadata[name]
            self._save_metadata(path, metadata)

    def _update_metadata_entry(self, path, name, data):
        with self._get_metadata_lock(path):
            metadata = copy.copy(self._get_metadata(path))
            metadata[name] = data
            self._save_metadata(path, metadata)

    def _copy_metadata_entry(
        self,
        source_path,
        source_name,
        destination_path,
        destination_name,
        delete_source=False,
        updates=None,
    ):
        with self._get_metadata_lock(source_path):
            source_data = self._get_metadata_entry(source_path, source_name, default={})
            if not source_data:
                return

            if delete_source:
                self._remove_metadata_entry(source_path, source_name)

        if updates is not None:
            source_data.update(updates)

        with self._get_metadata_lock(destination_path):
            self._update_metadata_entry(destination_path, destination_name, source_data)

    def _get_metadata(self, path, force=False):
        import json

        if not force:
            metadata = self._metadata_cache.get(path)
            if metadata:
                return metadata

        self._migrate_metadata(path)

        metadata_path = os.path.join(path, ".metadata.json")

        metadata = None
        with self._get_persisted_metadata_lock(path):
            if os.path.exists(metadata_path):
                with open(metadata_path, encoding="utf-8") as f:
                    try:
                        metadata = json.load(f)
                    except Exception:
                        self._logger.exception(
                            f"Error while reading .metadata.json from {path}"
                        )

        def valid_json(value):
            try:
                json.dumps(value, allow_nan=False)
                return True
            except Exception:
                return False

        if isinstance(metadata, dict):
            old_size = len(metadata)
            metadata = {k: v for k, v in metadata.items() if valid_json(v)}
            metadata = {
                k: v for k, v in metadata.items() if os.path.exists(os.path.join(path, k))
            }
            new_size = len(metadata)
            if new_size != old_size:
                self._logger.info(
                    "Deleted {} stale or invalid entries from metadata for path {}".format(
                        old_size - new_size, path
                    )
                )
                self._save_metadata(path, metadata)
            else:
                with self._get_metadata_lock(path):
                    self._metadata_cache[path] = metadata
            return metadata
        else:
            return {}

    def _save_metadata(self, path, metadata):
        import json

        with self._get_metadata_lock(path):
            self._metadata_cache[path] = metadata

        with self._get_persisted_metadata_lock(path):
            metadata_path = os.path.join(path, ".metadata.json")
            try:
                with atomic_write(metadata_path, mode="wb") as f:
                    f.write(
                        to_bytes(json.dumps(metadata, indent=2, separators=(",", ": ")))
                    )
            except Exception:
                self._logger.exception(f"Error while writing .metadata.json to {path}")

    def _delete_metadata(self, path):
        with self._get_metadata_lock(path):
            if path in self._metadata_cache:
                del self._metadata_cache[path]

        with self._get_persisted_metadata_lock(path):
            metadata_files = (".metadata.json", ".metadata.yaml")
            for metadata_file in metadata_files:
                metadata_path = os.path.join(path, metadata_file)
                if os.path.exists(metadata_path):
                    try:
                        os.remove(metadata_path)
                    except Exception:
                        self._logger.exception(
                            f"Error while deleting {metadata_file} from {path}"
                        )

    @staticmethod
    def _copied_metadata(metadata, name):
        metadata = copy.copy(metadata)
        metadata[name] = copy.deepcopy(metadata.get(name, {}))
        return metadata

    def _migrate_metadata(self, path):
        # we switched to json in 1.3.9 - if we still have yaml here, migrate it now
        import json

        with self._get_persisted_metadata_lock(path):
            metadata_path_yaml = os.path.join(path, ".metadata.yaml")
            metadata_path_json = os.path.join(path, ".metadata.json")

            if not os.path.exists(metadata_path_yaml):
                # nothing to migrate
                return

            if os.path.exists(metadata_path_json):
                # already migrated
                try:
                    os.remove(metadata_path_yaml)
                except Exception:
                    self._logger.exception(
                        f"Error while removing .metadata.yaml from {path}"
                    )
                return

            try:
                metadata = yaml.load_from_file(path=metadata_path_yaml)
            except Exception:
                self._logger.exception(f"Error while reading .metadata.yaml from {path}")
                return

            if not isinstance(metadata, dict):
                # looks invalid, ignore it
                return

            with atomic_write(metadata_path_json, mode="wb") as f:
                f.write(to_bytes(json.dumps(metadata, indent=2, separators=(",", ": "))))

            try:
                os.remove(metadata_path_yaml)
            except Exception:
                self._logger.exception(f"Error while removing .metadata.yaml from {path}")

    @contextmanager
    def _get_metadata_lock(self, path):
        with self._metadata_lock_mutex:
            if path not in self._metadata_locks:
                import threading

                self._metadata_locks[path] = (0, threading.RLock())

            counter, lock = self._metadata_locks[path]
            counter += 1
            self._metadata_locks[path] = (counter, lock)

        yield lock

        with self._metadata_lock_mutex:
            counter = self._metadata_locks[path][0]
            counter -= 1
            if counter <= 0:
                del self._metadata_locks[path]
            else:
                self._metadata_locks[path] = (counter, lock)

    @contextmanager
    def _get_persisted_metadata_lock(self, path):
        with self._persisted_metadata_lock_mutex:
            if path not in self._persisted_metadata_locks:
                import threading

                self._persisted_metadata_locks[path] = (0, threading.RLock())

            counter, lock = self._persisted_metadata_locks[path]
            counter += 1
            self._persisted_metadata_locks[path] = (counter, lock)

        yield lock

        with self._persisted_metadata_lock_mutex:
            counter = self._persisted_metadata_locks[path][0]
            counter -= 1
            if counter <= 0:
                del self._persisted_metadata_locks[path]
            else:
                self._persisted_metadata_locks[path] = (counter, lock)



__author__ = "Gina Häußge <gina@octoprint.org>"
__license__ = "GNU Affero General Public License http://www.gnu.org/licenses/agpl.html"
__copyright__ = "Copyright (C) 2017 The OctoPrint Project - Released under terms of the AGPLv3 License"

import gzip
import logging
import os
import re
import threading
from urllib import parse as urlparse

import webassets.filter.cssrewrite.urlpath as urlpath
from webassets.bundle import Bundle
from webassets.filter import Filter
from webassets.filter.cssrewrite.base import PatternRewriter
from webassets.filter.rjsmin import rjsmin
from webassets.merge import BaseHunk, MemoryHunk
from webassets.version import Manifest


def replace_url(source_url, output_url, url):
    # If path is an absolute one, keep it
    parsed = urlparse.urlparse(url)

    if not parsed.scheme and not parsed.path.startswith("/"):
        abs_source_url = urlparse.urljoin(source_url, url)

        # relpath() will not detect this case
        if urlparse.urlparse(abs_source_url).scheme:
            return abs_source_url

        # rewritten url: relative path from new location (output)
        # to location of referenced file (source + current url)
        url = urlpath.relpath(output_url, abs_source_url)

    return url


class UrlRewriter(PatternRewriter):
    def input(self, _in, out, **kw):
        source, source_path, output, output_path = (
            kw["source"],
            kw["source_path"],
            kw["output"],
            kw["output_path"],
        )

        self.source_path = source_path
        self.output_path = output_path
        self.source_url = self.ctx.resolver.resolve_source_to_url(
            self.ctx, source_path, source
        )
        self.output_url = self.ctx.resolver.resolve_output_to_url(self.ctx, output)

        return super().input(_in, out, **kw)

    def replace_url(self, url):
        return replace_url(self.source_url, self.output_url, url)


class LessImportRewrite(UrlRewriter):
    name = "less_importrewrite"

    patterns = {"import_rewrite": re.compile(r"(@import(\s+\(.*\))?\s+)\"(.*)\";")}

    def import_rewrite(self, m):
        import_with_options = m.group(1)
        import_url = self.replace_url(m.group(3))
        return f'{import_with_options}"{import_url}";'


class SourceMapRewrite(UrlRewriter):
    name = "sourcemap_urlrewrite"

    patterns = {"url_rewrite": re.compile(r"(//#\s+sourceMappingURL=)(.*)")}

    def url_rewrite(self, m):
        mapping = m.group(1)
        source_url = self.replace_url(m.group(2))
        return f"{mapping}{source_url}"


class SourceMapRemove(PatternRewriter):
    name = "sourcemap_remove"

    patterns = {"sourcemap_remove": re.compile(r"(//#\s+sourceMappingURL=)(.*)")}

    def sourcemap_remove(self, m):
        return ""


class JsDelimiterBundler(Filter):
    name = "js_delimiter_bundler"
    options = {}

    def input(self, _in, out, **kwargs):
        source = kwargs.get("source", "n/a")

        out.write("// source: " + source + "\n")
        out.write(_in.read())
        out.write("\n;\n")


class RJSMinExtended(Filter):
    name = "rjsmin_extended"
    options = {
        "keep_bang_comments": "RJSMIN_KEEP_BANG_COMMENTS",
    }

    def input(self, _in, out, **kw):
        source = kw["source"]
        if source.endswith(".min.js"):
            # out.write("// source is already minified, not minifying again\n")
            out.write(_in.read())
        else:
            keep = self.keep_bang_comments or False
            out.write(rjsmin.jsmin(_in.read(), keep_bang_comments=keep))


class GzipFile(Filter):
    name = "gzip"
    options = {}

    def output(self, _in, out, **kwargs):
        data = _in.read()
        out.write(data)

        # webassets requires us to output a "str", but we can't do that since gzip
        # provides binary outputs.
        #
        # We work around that by outputting the gzipped file to another path
        output_path = kwargs.get("output_path", None)
        if output_path:
            gzipped_output_path = output_path + ".gz"
            try:
                with gzip.open(gzipped_output_path, "wb", 9) as f:
                    f.write(data.encode("utf8"))
            except Exception:
                logging.getLogger(__name__).exception(
                    "Error writing gzipped "
                    "output of {} to {}".format(output_path, gzipped_output_path)
                )
                try:
                    os.remove(gzipped_output_path)
                except Exception:
                    logging.getLogger(__name__).exception(
                        "Error removing broken " ".gz from {}".format(gzipped_output_path)
                    )


class ChainedHunk(BaseHunk):
    def __init__(self, *hunks):
        self._hunks = hunks

    def mtime(self):
        pass

    def data(self):
        result = ""
        for hunk in self._hunks:
            if isinstance(hunk, tuple) and len(hunk) == 2:
                hunk, f = hunk
            else:
                f = lambda x: x
            result += f(hunk.data())
        return result


_PLUGIN_BUNDLE_WRAPPER_PREFIX = """// JS assets for plugin {plugin}
(function () {{
    try {{
        """
_PLUGIN_BUNDLE_WRAPPER_SUFFIX = """
    }} catch (error) {{
        log.error("Error in JS assets for plugin {plugin}:", `${{exc.message}}\n${{exc.stack || exc}}`);
    }}
}})();
"""


class JsPluginBundle(Bundle):
    def __init__(self, plugin, *args, **kwargs):
        Bundle.__init__(self, *args, **kwargs)
        self.plugin = plugin

    def _merge_and_apply(
        self,
        ctx,
        output,
        force,
        parent_debug=None,
        parent_filters=None,
        extra_filters=None,
        disable_cache=None,
    ):
        hunk = Bundle._merge_and_apply(
            self,
            ctx,
            output,
            force,
            parent_debug=parent_debug,
            parent_filters=parent_filters,
            extra_filters=extra_filters,
            disable_cache=disable_cache,
        )

        return ChainedHunk(
            MemoryHunk(_PLUGIN_BUNDLE_WRAPPER_PREFIX.format(plugin=self.plugin)),
            (hunk, lambda x: x.replace("\n", "\n        ")),
            MemoryHunk(_PLUGIN_BUNDLE_WRAPPER_SUFFIX.format(plugin=self.plugin)),
        )


class MemoryManifest(Manifest):
    id = "memory"

    def __init__(self, *args, **kwargs):
        self.manifest = {}
        self._lock = threading.RLock()

    def remember(self, bundle, ctx, version):
        with self._lock:
            self.manifest[bundle.output] = version

    def query(self, bundle, ctx):
        with self._lock:
            return self.manifest.get(bundle.output, None)



__license__ = "GNU Affero General Public License http://www.gnu.org/licenses/agpl.html"
__copyright__ = "Copyright (C) 2019 The OctoPrint Project - Released under terms of the AGPLv3 License"


import logging

import click

from octoprint import init_settings
from octoprint.access.groups import FilebasedGroupManager
from octoprint.access.users import (
    CorruptUserStorage,
    FilebasedUserManager,
    InvalidUsername,
    UnknownUser,
    UserAlreadyExists,
)
from octoprint.cli import get_ctx_obj_option
from octoprint.util import get_class, sv

click.disable_unicode_literals_warning = True

# ~~ "octoprint user" commands


@click.group()
@click.pass_context
def cli(ctx):
    """
    User management.

    Note that this currently only supports managing user accounts stored in the configured user manager, not any
    user managers added through plugins and the "octoprint.users.factory" hook.
    """
    try:
        logging.basicConfig(
            level=logging.DEBUG
            if get_ctx_obj_option(ctx, "verbosity", 0) > 0
            else logging.WARN
        )
        settings = init_settings(
            get_ctx_obj_option(ctx, "basedir", None),
            get_ctx_obj_option(ctx, "configfile", None),
            overlays=get_ctx_obj_option(ctx, "overlays", None),
        )

        group_manager_name = settings.get(["accessControl", "groupManager"])
        try:
            clazz = get_class(group_manager_name)
            group_manager = clazz()
        except AttributeError:
            click.echo(
                "Could not instantiate group manager {}, "
                "falling back to FilebasedGroupManager!".format(group_manager_name),
                err=True,
            )
            group_manager = FilebasedGroupManager()

        ctx.obj.group_manager = group_manager

        name = settings.get(["accessControl", "userManager"])
        try:
            clazz = get_class(name)
            user_manager = clazz(group_manager=group_manager, settings=settings)
        except CorruptUserStorage:
            raise
        except Exception:
            click.echo(
                "Could not instantiate user manager {}, falling back to FilebasedUserManager!".format(
                    name
                ),
                err=True,
            )
            user_manager = FilebasedUserManager(group_manager, settings=settings)

        ctx.obj.user_manager = user_manager

    except Exception:
        click.echo("Could not instantiate user manager", err=True)
        ctx.exit(-1)


@cli.command(name="list")
@click.pass_context
def list_users_command(ctx):
    """Lists user information"""
    users = ctx.obj.user_manager.get_all_users()
    _print_list(users)


@cli.command(name="add")
@click.argument("username", type=click.STRING, required=True)
@click.password_option("--password", "password", help="Password for the user")
@click.option("-g", "--group", "groups", multiple=True, help="Groups to set on the user")
@click.option(
    "-p",
    "--permission",
    "permissions",
    multiple=True,
    help="Individual permissions to set on the user",
)
@click.option(
    "--admin",
    "is_admin",
    type=click.BOOL,
    is_flag=True,
    default=False,
    help="Adds user to admin group",
)
@click.pass_context
def add_user_command(ctx, username, password, groups, permissions, is_admin):
    """Add a new user."""
    if not groups:
        groups = []

    if is_admin:
        groups.append(ctx.obj.group_manager.admin_group)

    try:
        ctx.obj.user_manager.add_user(
            username, password, groups=groups, permissions=permissions, active=True
        )

        user = ctx.obj.user_manager.find_user(username)
        if user:
            click.echo("User created:")
            click.echo(f"\t{_user_to_line(user.as_dict())}")
    except UserAlreadyExists:
        click.echo(f"A user with the name {username} does already exist!", err=True)
    except InvalidUsername:
        click.echo(f"The username '{username}' is invalid!", err=True)


@cli.command(name="remove")
@click.argument("username", type=click.STRING)
@click.pass_context
def remove_user_command(ctx, username):
    """Remove an existing user."""
    confirm = click.prompt(
        "This is will irreversibly destroy the user account! Enter 'yes' to confirm",
        type=click.STRING,
    )

    if confirm.lower() == "yes":
        ctx.obj.user_manager.remove_user(username)
        click.echo(f"User {username} removed.")
    else:
        click.echo(f"User {username} not removed.")


@cli.command(name="password")
@click.argument("username", type=click.STRING)
@click.password_option("--password", "password", help="New password for user")
@click.pass_context
def change_password_command(ctx, username, password):
    """Change an existing user's password."""
    try:
        ctx.obj.user_manager.change_user_password(username, password)
        click.echo(f"Password changed for user {username}.")
    except UnknownUser:
        click.echo(f"User {username} does not exist!", err=True)


@cli.command(name="activate")
@click.argument("username", type=click.STRING)
@click.pass_context
def activate_command(ctx, username):
    """Activate a user account."""
    try:
        ctx.obj.user_manager.change_user_activation(username, True)
        click.echo(f"User {username} activated.")

        user = ctx.obj.user_manager.find_user(username)
        if user:
            click.echo("User created:")
            click.echo(f"\t{_user_to_line(user.asDict())}")
    except UnknownUser:
        click.echo(f"User {username} does not exist!", err=True)


@cli.command(name="deactivate")
@click.argument("username", type=click.STRING)
@click.pass_context
def deactivate_command(ctx, username):
    """Activate a user account."""
    try:
        ctx.obj.user_manager.change_user_activation(username, False)
        click.echo(f"User {username} activated.")

        user = ctx.obj.user_manager.find_user(username)
        if user:
            click.echo("User created:")
            click.echo(f"\t{_user_to_line(user.asDict())}")
    except UnknownUser:
        click.echo(f"User {username} does not exist!", err=True)


def _print_list(users):
    click.echo(f"{len(users)} users registered in the system:")
    for user in sorted(
        map(lambda x: x.as_dict(), users), key=lambda x: sv(x.get("name"))
    ):
        click.echo(f"\t{_user_to_line(user)}")


def _user_to_line(user):
    return (
        "{name}"
        "\n\t\tactive: {active}"
        "\n\t\tgroups: {groups}"
        "\n\t\tpermissions: {permissions}".format(
            name=user.get("name"),
            active=user.get("active", "False"),
            groups=", ".join(user.get("groups", [])),
            permissions=", ".join(user.get("permissions", [])),
        )
    )



import logging
import urllib.parse


class TornadoAccessFilter(logging.Filter):
    def filter(self, record):
        try:
            status, request_line, rtt = record.args

            if status == 409:
                _, url, _ = request_line.split()
                u = urllib.parse.urlparse(url)
                if u.path in ("/api/printer",):
                    record.levelno = logging.INFO
                    record.levelname = logging.getLevelName(record.levelno)
        except Exception:
            logging.getLogger(__name__).exception(
                f"Error while filtering log record {record!r}"
            )

        return logging.Filter.filter(self, record)
